/**********************************************************************
 * $Id: cgienv2file.cpp,v 1.3 1999/05/12 19:01:48 daniel Exp $
 *
 * Name:     cgienv2file.cpp
 * Project:  Open Web Toolkit
 * Language: C++
 * Purpose:  CGI App. that captures the CGI environment and writes it
 *           to a file that can be used later on for debugging.
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1998, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: cgienv2file.cpp,v $
 * Revision 1.3  1999/05/12 19:01:48  daniel
 * Fixed problem with 'errno' parameter name in MyErrorHandler()
 *
 * Revision 1.2  1998/12/09 02:08:00  daniel
 * Converted to use portability library
 *
 * Revision 1.1  1998/12/09 01:44:41  daniel
 * Initial revision
 *
 **********************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "cpl_error.h"
#include "owtcgien.h"

extern "C" void MyErrorHandler(CPLErr nErrClass, int nErrno, const char *msg)
{
    fprintf(stdout, "Content-type: text/plain\n\n");
    fprintf(stdout, "ERROR %d: %s\n", nErrno, msg);

    exit(1);
}

/**********************************************************************
 *                       main()
 *
 ********************************************************************** */

int main() 
{
    /* Allocate a OWTCGIEnv object that will be used to handle CGI stuff.
     * When the program exits the object will be destroyed and any 
     * resources it uses will be freed.
     */
    OWTCGIEnv oCGIEnv;
#ifdef WIN32
    char      szOutFile[] = "\\temp\\cgienv.env";
#else
    char      szOutFile[] = "/tmp/cgienv.env";
#endif

    CPLSetErrorHandler(MyErrorHandler);

    /* Save CGI Env. State for debugging */
    oCGIEnv.DumpToFile( szOutFile );

    if (CPLGetLastErrorNo() == 0)
    {
        printf("Content-Type: text/plain\n\n");
        printf("CGI Environment saved to file %s\n", szOutFile);
    }

    return 0;
}

