/* 
 *  gificon.c
 *
 *  Return a rectangular GIF color sample.
 *
 *  The parameters for the color sample ot return are taken from the 
 *  PATH_INFO:
 *
 *  /cgi-bin/gificon/<width>/<height>/<R>/<G>/<B>
 *
 */

#include "cpl_string.h"
#include "gd.h"

#ifdef WIN32
#include <io.h>     /* _setmode()   */
#include <fcntl.h>  /* O_BINARY     */
#endif

int CreateGIFSample(FILE *fpOut, int nWidth, int nHeight, 
                    int nR, int nG, int nB);

int main() 
{
    int     nWidth=50, nHeight=50, nR=0, nG=0, nB=0;
    FILE    *fpOut;
    char    *pszPathInfo;

    printf("Content-Type: image/gif\n\n");

    pszPathInfo = getenv("PATH_INFO");

    fpOut = stdout;

    /* ----------------------------------------------------------
     * Read the PATH_INFO, and extract the values for size and color
     * ---------------------------------------------------------- */
    if (pszPathInfo && strlen(pszPathInfo) > 0)
    {
        char **papszParams;
        int  nParams;

        papszParams = CSLTokenizeStringComplex(pszPathInfo, "/", FALSE, FALSE);
        nParams = CSLCount(papszParams);

        if (nParams >= 0) nWidth  = atoi(papszParams[0]);
        if (nParams >= 1) nHeight = atoi(papszParams[1]);
        if (nParams >= 2) nR = atoi(papszParams[2]);
        if (nParams >= 3) nG = atoi(papszParams[3]);
        if (nParams >= 4) nB = atoi(papszParams[4]);

        CSLDestroy(papszParams);

        nWidth = MAX(nWidth, 2);
        nHeight = MAX(nHeight, 2);
    }


    /* ----------------------------------------------------------
     * Output the image 
     * (On WIN32 we have to make sure output file is in binary mode.)
     * ---------------------------------------------------------- */
#ifdef WIN32
    if ( _setmode( _fileno( fpOut ), O_BINARY ) == -1)
        CPLError(CE_Failure, CPLE_FileIO, "OWTChart::SaveSlideGIF(): "
                      "Cannot set output file mode (%s)",  strerror(errno));
#endif

    CreateGIFSample(fpOut, nWidth, nHeight, nR, nG, nB);
    
    return 0;
}


/**********************************************************************
 *                      CreateGIFSample()
 *
 * Produce the GIF color sample image and return it through fpOut.
 **********************************************************************/
int CreateGIFSample(FILE *fpOut, int nWidth, int nHeight, 
                    int nR, int nG, int nB)
{
    gdImagePtr im;
    int nColor;

    /* Allocate image in memory */
    im = gdImageCreate(nWidth, nHeight);

    /* Allocate first color... will become the bg color */
    nColor = gdImageColorAllocate(im, nR, nG, nB);

    /* Output the image */
    gdImageGif(im, fpOut);

    /* Destroy image in memory */
    gdImageDestroy(im);

    return 0;
}
