/* GDCHART 2  GDC.C 
 *
 * This version of the GDChart library was based on GDCHART 0.94b
 * (12 Nov. 1998) by Bruce Verderaime (brv@fred.net)
 *
 * Here is the original copyright notice:
 *
 +-----------------------------------------------------------------------
 | GDChart is free for use in your applications and for chart generation.
 | YOU MAY NOT re-distribute or represent the code as your own.
 | Any re-distributions of the code MUST reference the author, and include
 | any and all original documentaion.
 +-----------------------------------------------------------------------
 *
 * $Log: gdc.c,v $
 * Revision 1.5  2002/02/11 19:51:10  julien
 * Add Combo_Line_HLC and 3DCombo_Line_HLC type
 *
 * Revision 1.4  2002/01/11 15:51:16  julien
 * Add Combo_HLC_Line and 3DCombo_HLC_Line type
 *
 * Revision 1.3  2001/02/16 15:47:58  daniel
 * Added pie_requested_radius param.
 *
 * Revision 1.2  2000/11/20 19:58:13  daniel
 * Integrate changes between v0.94b and v0.10.1dev
 *
 */


#define _GDC_C_
#include "gdc.h"    /* Note: External programs should include gdchart.h only */

#include <string.h>

struct GDC_FONT_T GDC_fontc[GDC_numfonts] = { { (gdFontPtr)NULL, 8,  5 },
                                              { (gdFontPtr)NULL, 8,  5 },
                                              { (gdFontPtr)NULL, 12, 6 },
                                              { (gdFontPtr)NULL, 13, 7 },
                                              { (gdFontPtr)NULL, 16, 8 },
                                              { (gdFontPtr)NULL, 15, 9 } };

/* ------------------------------------------------------------------- *\ 
 * convert from enum GDC_font_size to gd fonts
 * for now load them all
 *  #defines and #ifdefs might enable loading only needed fonts
\* ------------------------------------------------------------------- */
void
GDC_load_font_conversions()
{
    GDC_fontc[GDC_pad].f     = gdFontTiny;
    GDC_fontc[GDC_TINY].f    = gdFontTiny;
    GDC_fontc[GDC_SMALL].f   = gdFontSmall;
    GDC_fontc[GDC_MEDBOLD].f = gdFontMediumBold;
    GDC_fontc[GDC_LARGE].f   = gdFontLarge;
    GDC_fontc[GDC_GIANT].f   = gdFontGiant;
}

/* ------------------------------------------------------------------ *\ 
 * count (natural) substrings (new line sep)
\* ------------------------------------------------------------------ */
short
GDC_cnt_nl( char    *nstr,
            int     *len )          /* strlen - max seg */
{
    short   c           = 1;
    short   max_seg_len = 0;
    short   tmplen      = 0;

    if( !nstr )
        {
        if( len )
            *len = 0;
        return 0;
        }

    while( *nstr )
        {
        if( *nstr == '\n' )
            {
            ++c;
            max_seg_len = MAX( tmplen, max_seg_len );
            tmplen = 0;
            }
        else
            ++tmplen;
        ++nstr;
        }

    if( len )
        *len = MAX( tmplen, max_seg_len );      /* don't forget last seg */
    return c;
}

/* ------------------------------------------------------------------ *\ 
 * gd out a string with '\n's
\* ------------------------------------------------------------------ */
void
GDCImageStringNL( gdImagePtr        im,
                  struct GDC_FONT_T *f,
                  int               x,
                  int               y,
                  char              *str,
                  int               clr,
                  GDC_JUSTIFY_T     justify )
{
    int     i;
    int     len;
    int     max_len;
    short   strs_num = GDC_cnt_nl( str, &max_len );
    //char    sub_str[max_len+1];
    DEF_ARRAY(char, sub_str, max_len+1);
    ALLOC_ARRAY(char, sub_str, max_len+1);

    len      = -1;
    strs_num = -1;
    i = -1;
    do
        {
        ++i;
        ++len;
        sub_str[len] = *(str+i);
        if( *(str+i) == '\n' ||
            *(str+i) == '\0' )
            {
            int xpos;

            sub_str[len] = '\0';
            ++strs_num;
            switch( justify )
              {
              case GDC_JUSTIFY_LEFT:    xpos = x;                    break;
              case GDC_JUSTIFY_RIGHT:   xpos = x+f->w*(max_len-len); break;
              case GDC_JUSTIFY_CENTER:
              default:                  xpos = x+f->w*(max_len-len)/2;
              }
            gdImageString( im,
                           f->f,
                           xpos,
                           y + (f->h-1)*strs_num,
                           (unsigned char *)sub_str,
                           clr );
            len = -1;
            }
        }
    while( *(str+i) );

    FREE_ARRAY(sub_str);
}

/* ------------------------------------------------------------------------ *\ 
 * todo:                                                                    *
 * really get a unique color from the color map                             *
\* ------------------------------------------------------------------------ */
long
get_uniq_color( gdImagePtr	im )
{
	return 0x123454;
}

/**********************************************************************
 *                       GDC_destroy_image()
 *
 * Safe replacement for gdImageDestroy()
 **********************************************************************/
void
GDC_destroy_image(void *im)
{
    if( im )
        gdImageDestroy( (gdImagePtr)im );
}

/**********************************************************************
 *                       GDC_out_err()
 *
 * Output an error message to the chart slide.
 **********************************************************************/
void
GDC_out_err( GDCChartPtr    psGDC,
             int            IMGWIDTH,
             int            IMGHEIGHT,
             FILE           *img_fptr,
             GDC_RGB_T      BGColor,
             GDC_RGB_T      LineColor,
             char           *err_str )
{

    gdImagePtr  im;
    int         lineclr;
    int         bgclr;


    if( (psGDC->hold_img & GDC_REUSE_IMAGE) &&
        psGDC->image != (void*)NULL )
        im = (gdImagePtr)psGDC->image;
    else
        im = gdImageCreate( IMGWIDTH, IMGHEIGHT );

    bgclr    = gdImageColorAllocate( im, l2gdcal(BGColor) );
    lineclr  = gdImageColorAllocate( im, l2gdcal(LineColor) );

    gdImageString( im,
                   gdFontMediumBold,
                   IMGWIDTH/2 - GDC_fontc[GDC_MEDBOLD].w*strlen(err_str)/2,
                   IMGHEIGHT/3,
                   (unsigned char *)err_str,
                   lineclr );

    /* usually psGDC->generate_img is used in conjunction with hard or hold options */
    if( psGDC->generate_img )
        {
        fflush(img_fptr);           // clear anything buffered 
        switch( psGDC->image_type )
            {
#ifdef HAVE_PNG
            case GDC_PNG:   gdImagePng( im, img_fptr );                     break;
#endif
#ifdef HAVE_JPEG
            case GDC_JPEG:  gdImageJpeg( im, img_fptr, psGDC->jpeg_quality );  break;
#endif
#ifdef HAVE_WBMP
            case GDC_WBMP:  gdImageWBMP( im, lineclr, img_fptr );         break;
#endif
            case GDC_GIF:
            default:        gdImageGif( im, img_fptr);
            }
        }

    if( psGDC->hold_img & GDC_EXPOSE_IMAGE )
        psGDC->image = (void*)im;
    else
        gdImageDestroy(im);
    return;
}


/**********************************************************************
 *                       GDCChartAlloc()
 *
 * Alloc a GDCChart structure and initialize it with the default values.
 *
 * Returns the new GDCChartPtr, or NULL in case of error.
 *
 * GDChartDestroy() should eventually be called to safely release the
 * memory used by the structure. 
 **********************************************************************/
GDCChartPtr GDCChartAlloc(GDC_CHART_T type, int imgwidth, int imgheight)
{
    GDCChartPtr psGDC;

    /*-----------------------------------------------------------------
     * Allocate a new structure
     *----------------------------------------------------------------*/
    psGDC = (GDCChartPtr)calloc(1, sizeof(GDCChart));

    if (psGDC == NULL)
    {
        /* __TODO__ We should produce an error message... 
         * why not use CPLError)() ???
         */
        fprintf(stderr, "Memory Allocation failed!\n");
        return NULL;
    }

    /*-----------------------------------------------------------------
     * Init struct's PRIVATE MEMBERS
     * These structure members are maintained by the library and should
     * not be set directly by the caller.
     *----------------------------------------------------------------*/
    psGDC->type         = type;
    psGDC->imgwidth     = imgwidth;
    psGDC->imgheight    = imgheight;

    psGDC->num_points   = 0;
    psGDC->num_sets     = 0;
    psGDC->num_groups   = 0;
    psGDC->num_hlc_sets = 0;
    psGDC->num_h2_sets  = 0;
    psGDC->num_fbar_sets= 0;
    psGDC->xlbl         = NULL;
    psGDC->uvals        = NULL;
    psGDC->uvals_allocated = FALSE;
    psGDC->uvol         = NULL;

    /*-----------------------------------------------------------------
     *Set default values for PUBLIC MEMBERS
     *----------------------------------------------------------------*/

    /**** COMMON OPTIONS ********************************/

    psGDC->image_type   = GDC_GIF;
    psGDC->jpeg_quality = -1;   /* 0-95 */
    psGDC->generate_img = TRUE;

    psGDC->hold_img     = GDC_DESTROY_IMAGE;
    psGDC->image        = (void*)NULL;

    /**** XY chart options *****************************/
    psGDC->ytitle       = NULL;
    psGDC->xtitle       = NULL;
    psGDC->ytitle2      = NULL;         /* ostesibly: volume label */
    psGDC->title        = NULL;         /* NLs ok here */
    psGDC->ltitle       = NULL;
    psGDC->rtitle       = NULL;

    psGDC->title_size   = GDC_MEDBOLD;
    psGDC->ytitle_size  = GDC_MEDBOLD;
    psGDC->xtitle_size  = GDC_MEDBOLD;
    psGDC->yaxisfont_size = GDC_SMALL;
    psGDC->xaxisfont_size = GDC_SMALL;

    psGDC->ylabel_fmt   = NULL;         /* printf fmt'ing, e.g.: "%.2f" */
    psGDC->ylabel2_fmt  = NULL;         /* default: "%.0f" future: fractions */
    psGDC->xlabel_spacing = 5;          /* pixels  MAXSHORT means force all */
    psGDC->xlabel_ctl   = NULL;         /* num_points[] TRUE,FALSE */

    psGDC->ylabel_density = 80;         /* % */
    psGDC->interpolations = FALSE;      /* GDC_INTERP_VALUE in data */
    psGDC->requested_ymin = GDC_NOVALUE;
    psGDC->requested_ymax = GDC_NOVALUE;
    psGDC->requested_ymin2 = GDC_NOVALUE;
    psGDC->requested_ymax2 = GDC_NOVALUE;
    psGDC->requested_yinterval = GDC_NOVALUE;
    psGDC->_0Shelf      = TRUE;         /* if applicable */
    psGDC->grid         = GDC_TICK_LABELS;
    psGDC->ticks        = GDC_TICK_LABELS;
    psGDC->xaxis        = TRUE;
    psGDC->yaxis        = TRUE;
    psGDC->yaxis2       = TRUE;
    psGDC->yval_style   = TRUE;
    psGDC->yscale_type  = GDC_SCALE_LINEAR;
    psGDC->stack_type   = GDC_STACK_DEPTH;
    psGDC->_3d_depth    = 5.0;          /* % gif size */
    psGDC->_3d_angle    = 45;           /* 1-89 */
    psGDC->bar_width    = 75;           /* % (1-100) */

    psGDC->HLC_style    = GDC_HLC_CLOSE_CONNECTED;
    psGDC->HLC_cap_width = 25;          /* % (1-100) */

    psGDC->annotation   = (GDC_ANNOTATION_T*)NULL;
    psGDC->annotation_font = GDC_SMALL;

    psGDC->num_scatter_pts = 0;
    psGDC->scatter      = (GDC_SCATTER_T*)NULL;
    psGDC->thumbnail    = FALSE;
    psGDC->thumblabel   = NULL;
    psGDC->thumbval     = -MAXFLOAT;
    psGDC->border       = GDC_BORDER_ALL;

    psGDC->BGColor      = 0x000000L;    /* black */
    psGDC->GridColor    = 0xA0A0A0L;    /*  gray */
    psGDC->LineColor    = GDC_DFLTCOLOR;
    psGDC->PlotColor    = GDC_DFLTCOLOR;
    psGDC->VolColor     = 0xA0A0FFL;    /* lgtblue1 */
    psGDC->TitleColor   = GDC_DFLTCOLOR;/* "opposite" of BG */
    psGDC->XTitleColor  = GDC_DFLTCOLOR;
    psGDC->YTitleColor  = GDC_DFLTCOLOR;
    psGDC->YTitle2Color = GDC_DFLTCOLOR;
    psGDC->XLabelColor  = GDC_DFLTCOLOR;
    psGDC->YLabelColor  = GDC_DFLTCOLOR;
    psGDC->YLabel2Color = GDC_DFLTCOLOR;

    /* supercedes VolColor  ulong_color[num_points] */
    psGDC->ExtVolColor  = (GDC_RGB_T *)NULL;
    /* supercedes LineColor ulong_color[num_sets] */
    psGDC->SetColor     = (GDC_RGB_T *)NULL;
    /* supercedes SetColor  ulong_color[num_sets][num_points] */
    psGDC->ExtColor     = (GDC_RGB_T *)NULL;

    psGDC->LineStyles   = NULL;

    psGDC->transparent_bg = FALSE;
    psGDC->BGImage      = (char*)NULL;

    /* legends?  separate gif? */
    /* auto-size fonts, based on GIF size? */

    /* ----- following options are for expert users only ----- */
                        /* for alignment of multiple charts */
                        /* USE WITH CAUTION! */
    psGDC->hard_size        = FALSE;
    psGDC->hard_xorig       = 0;        /* in/out */
    psGDC->hard_graphwidth  = 0;        /* in/out */
    psGDC->hard_yorig       = 0;        /* in/out */
    psGDC->hard_grapheight  = 0;        /* in/out */

    /******************************************************/
    /****       OPTIONS SPECIFIC TO PIE CHARTS        *****/
    /******************************************************/
    psGDC->pie_EdgeColor        = GDC_NOCOLOR;
    psGDC->pie_other_threshold  = -1;
    psGDC->pie_label_size       = GDC_SMALL;
    psGDC->pie_label_dist       = 1+8/2; /* 1+GDC_fontc[pie_label_size].h/2 */
    psGDC->pie_label_line       = FALSE; /* from label to slice */
    psGDC->pie_explode          = NULL;  /* [num_points] */
    psGDC->pie_Color            = NULL;
    psGDC->pie_missing          = NULL;  /* TRUE/FALSE */
    psGDC->pie_percent_labels   = GDCPIE_PCT_NONE;
    psGDC->pie_percent_fmt      = "%.0f%%";
    psGDC->pie_requested_radius = GDC_NOVALUE;

    /*-----------------------------------------------------------------
     * Some options shared between XY and PIE charts have a different
     * default value depending on chart type.
     *----------------------------------------------------------------*/
    if (psGDC->type == GDC_3DPIE || psGDC->type == GDC_2DPIE)
    {
        psGDC->_3d_depth = 10;  /* % of GIF width */
    }

    /*-----------------------------------------------------------------
     * OK, we're done ... return new struct.
     *----------------------------------------------------------------*/
    return psGDC;
}


/**********************************************************************
 *                       GDCChartGIF()
 *
 * Draw the chart in the GIF image.  
 *
 * This function will call the right drawing function based on
 * chart type.
 *
 * little/no error checking  0:    ok,
 *                           -ret: error no graph output
 *                           ret:  error graph out
 *
 * No check for a gif that's too small to fit needed info (labels, etc), 
 * could core dump.
 * watch out for # params and array sizes==num_points
 **********************************************************************/
int  GDCChartGIF( GDCChartPtr psGDC,
                  FILE        *img_fptr)
{
    if (psGDC->type == GDC_3DPIE ||
        psGDC->type == GDC_2DPIE)
    {
        return _GDCPieChartGIF(psGDC, img_fptr);
    }
    else
    {
        return _GDCXYChartGIF(psGDC, img_fptr);
    }
}

/**********************************************************************
 *                       GDCChartDestroy()
 *
 * Release all memory used by a GDCChart structure.
 *
 * See Also: GDCChartAlloc()
 **********************************************************************/
void   GDCChartDestroy(GDCChartPtr psGDC)
{
    if (psGDC == NULL)
        return;

    /*-----------------------------------------------------------------
     * For now there is not much stuff to free (everything is the 
     * responsibility of the caller), but eventually, we should free 
     * arrays, strings, etc. here.
     *----------------------------------------------------------------*/
    if (psGDC->uvals != NULL && psGDC->uvals_allocated)
        free(psGDC->uvals);

    free(psGDC);
}


/**********************************************************************
 *                       GDCChartSetValues()
 *
 * Set the values to chart.  All the values are passed in one single array,
 * in the order defined below.  
 *
 * This function replaces GDC_set_values() which used var. args.
 * The problem with var. args is that you must know the number of sets
 * for the chart at compile time.  We bypass that limitation by passing
 * all the values in one single array that will be built dynamically
 * by the caller.
 *
 * The array of values passed to the function is not copied, we just keep
 * a reference to them.  So it will be up to the caller to release the
 * memory associated with it (if necessary), but ONLY AFTER the chart
 * has been generated.
 *
 * Expected order of the values in pafValues[] for each chart type:
 *     GDC_LINE
 *     GDC_BAR
 *     GDC_3DBAR
 *     GDC_3DAREA
 *     GDC_AREA    If num_sets = 1:
 *                   { val1, val2, val3 ... val_num_points }
 *
 *                 If num_sets > 1:
 *                   { val[1][1], val[1][2], val[1][3] ... val[1][num_points],
 *                    val[2][1], val[2][2], val[2][3] ... val[2][num_points],
 *                    ...
 *                    val[num_sets][1], ... val[num_sets][num_points] }
 *
 * multiple sets make sense for rest?
 *
 *     GDC_HILOCLOSE       { high_1, high2, high3, ...  high_num_points,
 *                           low_1,  low2,  low3, ...  low_num_points,
 *                           close_1, close2, close3, ... close_num_points }
 *     
 *     GDC_COMBO_LINE_BAR
 *     GDC_COMBO_LINE_LINE
 *     GDC_COMBO_LINE_AREA { val1, val2, val3 ... val_num_points,
 *                           vol1, vol2, vol3 ... vol_num_points }
 *     
 *     GDC_COMBO_HLC_BAR
 *     GDC_COMBO_HLC_AREA  { high_1, high2, high3, ...  high_num_points,
 *                           low_1,  low2,  low3, ...  low_num_points,
 *                           close_1, close2, close3, ... close_num_points,
 *                           vol1, vol2, vol3 ... vol_num_points }

 *     GDC_FLOATINGBAR     { low_1,  low2,  low3, ...  low_num_points,
 *                           high_1, high2, high3, ...  high_num_points )
 *
 *     GDC_3DPIE
 *     GDC_2DPIE    num_sets must be = 1
 *                   { val1, val2, val3 ... val_num_points }
 **********************************************************************/

int  GDCChartSetValues( GDCChartPtr psGDC,
                        int      num_points, /* points along x axis (even   */
                                             /* interval) all arrays        */
                                             /* dependant on this           */
                        char     *xlbl[],    /* array of xlabels            */
                        int      num_sets,
                        float    *pafValues )
{
    int         i;

    psGDC->num_points   = num_points;
    psGDC->num_groups   = num_sets;
    psGDC->num_sets     = num_sets;
    psGDC->num_hlc_sets = ( psGDC->type == GDC_COMBO_HLC_BAR   ||
                            psGDC->type == GDC_COMBO_HLC_AREA  ||
                            psGDC->type == GDC_COMBO_HLC_LINE  ||
                            psGDC->type == GDC_3DCOMBO_HLC_BAR ||
                            psGDC->type == GDC_3DCOMBO_HLC_AREA||
                            psGDC->type == GDC_3DCOMBO_HLC_LINE||
                            psGDC->type == GDC_3DHILOCLOSE     ||
                            psGDC->type == GDC_HILOCLOSE )? num_sets: 0;
    psGDC->num_h2_sets =  ( psGDC->type == GDC_H2BAR   ||
                            psGDC->type == GDC_3DH2BAR )? num_sets: 0;
    psGDC->num_fbar_sets =( psGDC->type == GDC_FLOATINGBAR   ||
                            psGDC->type == GDC_3DFLOATINGBAR )? num_sets: 0;

    psGDC->xlbl         = xlbl;

    /*-----------------------------------------------------------------
     * Alloc an array for the pointers to the user provided values.
     * Since this array is not provided by the user, we'll have to 
     * free it in GDCChartDestroy().
     *
     * psGDC->num_sets is adjusted to become the real number of sets of
     * values in uvals[].  psGDC->num_groups gives the original num_sets value.
     *----------------------------------------------------------------*/
    psGDC->num_sets = psGDC->num_hlc_sets?(psGDC->num_hlc_sets*3):
                      psGDC->num_h2_sets?(psGDC->num_h2_sets*2):
                      psGDC->num_fbar_sets?(psGDC->num_fbar_sets*2): num_sets;

    if (psGDC->uvals != NULL && psGDC->uvals_allocated)
        free(psGDC->uvals);
    psGDC->uvals = (float **)calloc(psGDC->num_sets, sizeof(float*));
    if (psGDC->uvals == NULL)
    {
        fprintf(stderr, "Memory Allocation failed\n");
        return 1;
    }
    psGDC->uvals_allocated = TRUE;

    /*-----------------------------------------------------------------
     * build array of sets (each set will be stored in an array of floats)
     *----------------------------------------------------------------*/
    for( i=0; i < psGDC->num_sets; i++ )
    {
        psGDC->uvals[i] = pafValues + i*num_points;
    }

    if( psGDC->type == GDC_COMBO_HLC_BAR   || /* aka: combo */
        psGDC->type == GDC_COMBO_HLC_AREA  ||
        psGDC->type == GDC_COMBO_HLC_LINE  ||
        psGDC->type == GDC_COMBO_LINE_HLC  ||
        psGDC->type == GDC_COMBO_LINE_BAR  ||
        psGDC->type == GDC_COMBO_LINE_AREA ||
        psGDC->type == GDC_COMBO_LINE_LINE ||
        psGDC->type == GDC_3DCOMBO_HLC_BAR ||
        psGDC->type == GDC_3DCOMBO_HLC_AREA||
        psGDC->type == GDC_3DCOMBO_HLC_LINE||
        psGDC->type == GDC_3DCOMBO_LINE_HLC||
        psGDC->type == GDC_3DCOMBO_LINE_BAR||
        psGDC->type == GDC_3DCOMBO_LINE_AREA||
        psGDC->type == GDC_3DCOMBO_LINE_LINE )
    {
        psGDC->uvol = pafValues + psGDC->num_sets*num_points;
    }
    else
    {
        psGDC->uvol = NULL;
    }

    /*-----------------------------------------------------------------
     * For PIE charts, GDC_NOVALUE is not supported, but just in case
     * some would have been received, set these values to 0.
     * ... and BTW, why not make it monkey-proof... flush all values < 0
     *----------------------------------------------------------------*/
    if( psGDC->type == GDC_2DPIE || psGDC->type == GDC_3DPIE )
    {
        for( i=0; i < num_points; i++ )
        {
            if (psGDC->uvals[0][i] == GDC_NOVALUE || psGDC->uvals[0][i] < 0.0)
                psGDC->uvals[0][i] = 0;
        }
    }

    return 0;
}
