/* GDCHART 2  GDC.H 
 *
 ************************************************************************
 * GDC.H contains stuff for internal library use.  It should not be included
 * by external programs.
 *
 * Programs that want to use the GDChart library should include GDCHART.H
 * instead.
 *
 ************************************************************************
 * This version of the GDChart library was based on GDCHART 0.94b
 * (12 Nov. 1998) by Bruce Verderaime (brv@fred.net)
 *
 * Here is the original copyright notice:
 *
 +-----------------------------------------------------------------------
 | GDChart is free for use in your applications and for chart generation.
 | YOU MAY NOT re-distribute or represent the code as your own.
 | Any re-distributions of the code MUST reference the author, and include
 | any and all original documentaion.
 +-----------------------------------------------------------------------
 *
 * $Log: gdc.h,v $
 * Revision 1.2  2000/11/20 19:58:13  daniel
 * Integrate changes between v0.94b and v0.10.1dev
 *
 */

#ifndef _GDC_H
#define _GDC_H

#include <stdlib.h>

#include "gdchart.h"    /* Include public definitions */

#include "gdfonts.h"
#include "gdfontt.h"
#include "gdfontmb.h"
#include "gdfontg.h"
#include "gdfontl.h"


#ifdef __cplusplus
extern "C" {
#endif

/*---------------------------------------------------------------------
 *  Font stuff
 *--------------------------------------------------------------------*/

struct  GDC_FONT_T  {
                    gdFontPtr   f;
                    char        h;
                    char        w;
                    };

typedef enum { GDC_JUSTIFY_RIGHT,
               GDC_JUSTIFY_CENTER,
               GDC_JUSTIFY_LEFT 
             } GDC_JUSTIFY_T;

#ifndef _GDC_C_
  extern struct   GDC_FONT_T  GDC_fontc[];
#endif

/*---------------------------------------------------------------------
 *  Internal library Function prototypes (gdc.c)
 *--------------------------------------------------------------------*/

void    GDC_load_font_conversions();
short   GDC_cnt_nl( char*, int* );
void    GDC_out_err( GDCChartPtr    psGDC,
                     int            GIFWIDTH,
                     int            GIFHEIGHT,
                     FILE*,
                     GDC_RGB_T      BGColor,
                     GDC_RGB_T      LineColor,
                     char           *str );

void    GDCImageStringNL( gdImagePtr, struct GDC_FONT_T*, int, int, char*, 
                          int, GDC_JUSTIFY_T );

char*   GDC_price_to_str( float price,
                          int   *numorator,
                          int   *demoninator,
                          int   *decimal,
                          char  *fltfmt );

int     _GDCXYChartGIF( GDCChartPtr psGDC, FILE *fpOut);
int     _GDCPieChartGIF( GDCChartPtr psGDC, FILE *fpOut);

/*---------------------------------------------------------------------
 *  Struct to hold patterned line style definitions (see gdImageSetStyle)
 *--------------------------------------------------------------------*/
typedef struct GDCLineStyle_t
{
    int *Style;
    int StyleLength;
    int LineColor;  /* Either a GD allocated color or 'gdStyled' */
}GDCLineStyle;

/*---------------------------------------------------------------------
 *  Macros to allocate colors
 *--------------------------------------------------------------------*/

#define DEF_CLR_ALLOCATE_VARS  \
   int            _gdccfoo1; \
   GDC_RGB_T      _gdccfoo2

#define _gdcntrst(bg)       ( ((bg)&0x800000?0x000000:0xFF0000)|    \
                              ((bg)&0x008000?0x000000:0x00FF00)|    \
                              ((bg)&0x000080?0x000000:0x0000FF) )
#define _clrallocate( im, rawclr, bgc )                                 \
        ( (_gdccfoo2=rawclr==GDC_DFLTCOLOR? _gdcntrst(bgc): rawclr),    \
          (_gdccfoo1=gdImageColorExact(im,l2gdcal(_gdccfoo2))) != -1?   \
            _gdccfoo1:                                                  \
            gdImageColorsTotal(im) == gdMaxColors?                      \
               gdImageColorClosest(im,l2gdcal(_gdccfoo2)):              \
               gdImageColorAllocate(im,l2gdcal(_gdccfoo2)) )

#define _clrshdallocate( im, rawclr, bgc )                              \
        ( (_gdccfoo2=rawclr==GDC_DFLTCOLOR? _gdcntrst(bgc): rawclr),    \
          (_gdccfoo1=gdImageColorExact(im,l2gdshd(_gdccfoo2))) != -1?   \
            _gdccfoo1:                                                  \
            gdImageColorsTotal(im) == gdMaxColors?                      \
                gdImageColorClosest(im,l2gdshd(_gdccfoo2)):             \
                gdImageColorAllocate(im,l2gdshd(_gdccfoo2)) )

#define clrallocate( im, rawclr )     \
                             _clrallocate( im, rawclr, psGDC->BGColor )
#define clrshdallocate( im, rawclr )  \
                            _clrshdallocate( im, rawclr, psGDC->BGColor )


/*---------------------------------------------------------------------
 * Since MSVC++ does not support non-static auto arrays, we will use
 * the following macros to dynamically allocate the arrays and we will
 * free them at the end of the functions
 *--------------------------------------------------------------------*/
/*  OLD VERSION ...
 *#define ALLOC_ARRAY(type, name, size)  \
 *               type *name; name = (type *)calloc((size), sizeof(type))
 *
 *#define ALLOC_2DARRAY(type, name, w, h) \
 *              type *name; int w_##name = (w); \
 *              name = (type *)calloc(((w)*(h)), sizeof(type))
 *..... */

#define DEF_ARRAY(type, name, size) \
               type *name = NULL
#define ALLOC_ARRAY(type, name, size)  \
               name = (type *)calloc((size), sizeof(type)); \
               if (name == NULL)  \
                   { fprintf(stderr, "Memory Allocation Failed!\n"); abort();}

#define DEF_2DARRAY(type, name, w, h) \
                type *name = NULL; int h_##name = (h)
#define ALLOC_2DARRAY(type, name, w, h) \
                name = (type *)calloc(((w)*(h)), sizeof(type)); \
                if (name == NULL)  \
                   { fprintf(stderr, "Memory Allocation Failed!\n"); abort();}

#define _2D(name, i, j) name[h_##name*(i)+(j)]
#define FREE_ARRAY(name)   free(name)


#ifdef __cplusplus
}
#endif

#endif /* _GDC_H_ */
