/**********************************************************************
 * $Id: htdoc.cpp,v 1.3 1999/06/09 19:15:25 daniel Exp $
 *
 * Name:     wmhtdoc.cpp
 * Project:  OWT HTML manipulation library
 * Purpose:  Implementation of the OWTHtDoc class.
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1998, 1999, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: htdoc.cpp,v $
 * Revision 1.3  1999/06/09 19:15:25  daniel
 * Fixed crash when LoadFromFile() called twice in the life of an OWTHtDoc
 *
 * Revision 1.2  1998/12/09 02:08:04  daniel
 * Converted to use portability library
 *
 * Revision 1.1  1998/12/04 19:15:17  daniel
 * Initial revision
 *
 *
 **********************************************************************/

#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "owtht.h"

/*=====================================================================
                        OWTHtDoc class.
 =====================================================================*/

/**********************************************************************
 *                       OWTHtDoc::OWTHtDoc()
 *
 **********************************************************************/
OWTHtDoc::OWTHtDoc()
            :OWTHtBlock("MainDoc", "Doc", NULL)
{
    /* Default Content-Type */
    m_pszContentType = strdup("text/html");
}

/**********************************************************************
 *                       OWTHtDoc::~OWTHtDoc()
 *
 **********************************************************************/
OWTHtDoc::~OWTHtDoc()
{
    if (m_pszContentType)
        free(m_pszContentType);
}

/**********************************************************************
 *                       OWTHtDoc::SetContentType()
 *
 **********************************************************************/
void OWTHtDoc::SetContentType(const char *pszContentType)
{
    if (m_pszContentType)
        free(m_pszContentType);

    m_pszContentType = strdup(pszContentType);
}

/**********************************************************************
 *                       OWTHtDoc::Draw()
 *
 **********************************************************************/
void OWTHtDoc::Draw(FILE *fpOut /*=NULL*/)
{
    if (fpOut == NULL)
        fpOut = stdout;

    /* Include a HTTP header, and call the OWTHtBlock Draw() method.
     */
    fprintf(fpOut, "Content-Type: %s\n\n", m_pszContentType);

    OWTHtBlock::Draw(fpOut);
}

/**********************************************************************
 *                       OWTHtDoc::LoadFromFile()
 *
 * Returns TRUE on success or FALSE if something wrong happened.
 **********************************************************************/
GBool    OWTHtDoc::LoadFromFile(const char *pszFname)
{
    char **papszText;

    papszText = CSLLoad(pszFname);

    SetText(papszText, FALSE);

    return (papszText != NULL);
}

/**********************************************************************
 *                       OWTHtDoc::SaveToFile()
 *
 * Returns TRUE on success or FALSE if something wrong happened.
 **********************************************************************/
GBool    OWTHtDoc::SaveToFile(const char *pszFname)
{
    FILE    *fp;

    if (m_papszText)
    {
        if ((fp = fopen(pszFname, "wt")) != NULL)
        {
            OWTHtBlock::Draw(fp);

            fclose(fp);
        }
        else
        {
            /* Unable to open file */
            CPLError(CE_Failure, CPLE_FileIO,
                     "SaveToFile(%s): %s", pszFname, strerror(errno));
            return FALSE;
        }
    }

    return TRUE;
}
