/**********************************************************************
 * $Id: owtcgien.h,v 1.5 2003/01/31 04:10:36 daniel Exp $
 *
 * Name:     owtcgien.h
 * Project:  OWT base library
 * Purpose:  Header file for the OWTCGIEnv class used to access 
 *           HTML FORM parameters and CGI env. vars
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1998-2003, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: owtcgien.h,v $
 * Revision 1.5  2003/01/31 04:10:36  daniel
 * Added OWTCGIEnv::InitFromQueryString() and modified OWTParseForm()
 * to accept QUERY_STRING parameter to bypass env. var.
 *
 * Revision 1.4  2000/11/29 20:29:26  daniel
 * Added OWTDecodeHTML2ASCII()
 *
 * Revision 1.3  1999/05/28 16:59:52  daniel
 * Added SetEnv(), and funcs to build URL and decode accentuated chars
 *
 * Revision 1.2  1998/12/09 02:08:01  daniel
 * Converted to use portability library
 *
 * Revision 1.1  1998/12/04 19:15:18  daniel
 * Initial revision
 *
 *
 **********************************************************************/

#ifndef _OWTCGIEN_H_INCLUDED_
#define _OWTCGIEN_H_INCLUDED_

#include "cpl_string.h"

/*=====================================================================
                            Class Definitions
 =====================================================================*/

/*---------------------------------------------------------------------
 *                        (class) OWTCGIEnv
 *
 * This class provides a simple interface to access the CGI FORM 
 * parameters and CGI environment variables.  
 * It automatically handles both POST and GET methods, etc.
 *--------------------------------------------------------------------*/
class OWTCGIEnv
{
protected:
    GBool       m_bParsed;
    char      **m_papszParams;
    char      **m_papszEnvVars;

    void        ParseForm();

public:
    OWTCGIEnv();
    ~OWTCGIEnv();

    const char *GetFormParam(const char *pszName);
    char      **GetFormParamMulti(const char *pszName);
    char      **GetAllFormParams();

    const char *GetEnv(const char *pszname);
    void        SetEnv(const char *pszName, const char *pszValue);

    void    DumpToFile(const char *pszFname);
    void    InitFromFile(const char *pszFname);
    void    InitFromQueryString(const char *pszQueryString);
};


/*=====================================================================
                            Public functions
 =====================================================================*/
char **OWTParseForm(const char *pszQueryString);
char *OWTBuildURL(const char *pszURLPath, char **papszParams);

void  OWTDecodeHTMLCharSet(char *pszStr, const char *pszCodes);
void  OWTDecodeHTML2ISOLatin1(char *pszStr);
void  OWTDecodeHTML2ASCII(char *pszStr);

#endif /* _OWTCGIEN_H_INCLUDED_ */
