/**********************************************************************
 * $Id: owtchart.h,v 1.4 1999/05/28 17:01:52 daniel Exp $
 *
 * Name:     owtchart.h
 * Project:  OWT - Chart building block
 * Purpose:  Header file for the OWTChart class.
 * Note:     This class uses the GifDraw (gd1.3) library for its rendering
 *           engine.
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1998, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: owtchart.h,v $
 * Revision 1.4  1999/05/28 17:01:52  daniel
 * GetLegendSample() now takes cgi sample program path as argument
 *
 * Revision 1.3  1999/01/06 14:10:13  daniel
 * Added virtual class OWTBaseChart
 *
 * Revision 1.2  1998/12/09 02:08:05  daniel
 * Converted to use portability library
 *
 * Revision 1.1  1998/12/04 19:15:19  daniel
 * Initial revision
 *
 * Revision 1.3  1998/11/06 03:50:11  daniel
 * First working version of the Charting Demo application
 *
 * Revision 1.2  1998/10/30 16:40:05  daniel
 * Added RCS header
 *
 **********************************************************************/

#ifndef _OWTCHART_H_INCLUDED_
#define _OWTCHART_H_INCLUDED_

#include "cpl_port.h"
#include "gd.h"


#define OWTCHART_NO_VALUE    -1e25


class OWTBaseChart
{
public:
    virtual void    SetSize(int nWidth, int nHeight) = 0;
    virtual void    SetValues(int numLines, int numValPerLine, double *pValues) = 0;
    virtual void    SetValues(int numLines, int numValPerLine, char **papszValues) = 0;

    virtual void    SetXValueLabels(char **papszValueLabels) = 0;
    virtual char    **GetLegendSamples(const char *pszGIFSampleURL,
                                       int nWidth, int nHeight) = 0;

    // virtual int     SaveSlideGIF(FILE *fpOut) = 0;
    // virtual int     SaveSlideGIF(const char *pszFname) = 0;

};

 
/*=====================================================================
                            OWTChart class
 =====================================================================*/

class OWTChart: public OWTBaseChart
{
private:
    int         m_nWidth;
    int         m_nHeight;

    int         m_numLines;
    int         m_numValPerLine;
    double      *m_pValues;

    double      m_dMinValue;
    double      m_dMaxValue;

    char        **m_papszXValueLabels;

    /* The following members are used during the Redraw() call */
    int         m_nXLeft;   // Pixel coords of the chart rectangle
    int         m_nXRight;
    int         m_nYTop;
    int         m_nYBottom;

    gdFontPtr   m_pFont;


    void    Redraw(gdImagePtr pImage);
    void    ChartValues(gdImagePtr pImage, int nColor);
    void    ChartAnnotations(gdImagePtr pImage, int nColor);

public:
    int         m_numYAxisDiv;
    GBool        m_bHorGrid;
    GBool        m_bYAxisTickMarks;
    GBool        m_bXAxisTickMarks;

    OWTChart();
    virtual ~OWTChart();

    void    SetSize(int nWidth, int nHeight);
    void    SetValues(int numLines, int numValPerLine, double *pValues);
    void    SetValues(int numLines, int numValPerLine, char **papszValues);

    void    SetXValueLabels(char **papszValueLabels);
    char    **GetLegendSamples(const char *pszGIFSampleURL,
                               int nWidth, int nHeight);

    int     SaveSlideGIF(FILE *fpOut);
    int     SaveSlideGIF(const char *pszFname);

};


#endif /* _OWTCHART_H_INCLUDED_ */
