/**********************************************************************
 * $Id: owtgdchart.h,v 1.3 2002/04/19 17:54:12 julien Exp $
 *
 * Name:     owtgdchart.h
 * Project:  OWT - Chart building block
 * Purpose:  Header file for the OWTGDChart class.
 * Note:     This class uses the GDChart and GifDraw (gd1.3) libraries..
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1998, 1999, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: owtgdchart.h,v $
 * Revision 1.3  2002/04/19 17:54:12  julien
 * Added GetParams function
 *
 * Revision 1.2  1999/07/27 20:10:58  daniel
 * Added possibility to have groups of lines the same color (for LCDC apps)
 *
 * Revision 1.1  1999/05/28 17:04:21  daniel
 * Initial revision
 *
 **********************************************************************/

#ifndef _OWTGDCHART_H_INCLUDED_
#define _OWTGDCHART_H_INCLUDED_

#include "cpl_port.h"
#include "gd.h"

#ifdef WIN32
#  pragma warning( once : 4244 ) /* Issue warning 4244 only once. */
#endif

#include "gdchart.h"

#include "owtchart.h"

/*=====================================================================
                            OWTGDChart class
 =====================================================================*/

class OWTGDChart: public OWTBaseChart
{
private:
    GDCChartPtr psGDC;

    int         m_numLines;
    int         m_numValPerLine;
    float      *m_pfValues;

    char        **m_papszXValueLabels;

    /* The following members are used during the Redraw() call */
    int         m_nXLeft;   // Pixel coords of the chart rectangle
    int         m_nXRight;
    int         m_nYTop;
    int         m_nYBottom;

    gdFontPtr   m_pFont;


    void    Redraw(FILE *fpOut);
    void    RedrawPie(FILE *fpOut);

public:
    int         m_numYAxisDiv;
    GBool        m_bHorGrid;
    GBool        m_bYAxisTickMarks;
    GBool        m_bXAxisTickMarks;

    OWTGDChart(GDC_CHART_T nChartType);
    virtual ~OWTGDChart();

    void    SetSize(int nWidth, int nHeight);
    void    SetValues(int numLines, int numValPerLine, double *pValues);
    void    SetValues(int numLines, int numValPerLine, char **papszValues);

    void    SetXValueLabels(char **papszValueLabels);
    char    **GetLegendSamples(const char *pszGIFSampleURL,
                               int nWidth, int nHeight);

    int     SaveSlideGIF(FILE *fpOut);
    int     SaveSlideGIF(const char *pszFname);

};


/*=====================================================================
 *                          OWTGDChartURL class
 *
 * Similar to OWTGDChart, except that it only produces a URL for the
 * OWTChart CGI engine, instead of producing a GIF slide.
 =====================================================================*/

class OWTGDChartURL: public OWTBaseChart
{
private:
    char        **m_papszParams;
    char        *m_pszURLBuf;

    GDC_CHART_T m_eType;
    int         m_numLines;
    int         m_numValPerLine;

    char        **m_papszStyles;
    int         m_numLinesByColor;

    void        Redraw();
    void        RedrawPie();
    char        **GetColorsArray(int numColors, int numLinesByColor = 1);

public:

    OWTGDChartURL(GDC_CHART_T nChartType);
    virtual     ~OWTGDChartURL();

    void        SetURLParam(const char *pszName, const char *pszValue);
    void        SetURLParam(const char *pszName, int nValue);
    void        SetURLParam(const char *pszName, char **papszValues,
                            char cDelim = ';');

    void        SetChartType(GDC_CHART_T nChartType);
    void        SetSize(int nWidth, int nHeight);
    void        SetValues(int numLines, int numValPerLine, double *pValues);
    void        SetValues(int numLines, int numValPerLine, char **papszValues);

    void        SetXValueLabels(char **papszValueLabels);
    void        SetTitle(const char *pszTitle);
    void        SetXAxisTitle(const char *pszTitle);
    void        SetYAxisTitle(const char *pszTitle);
    void        SetLeftRightTitles(const char *pszLTitle,
                                  const char *pszRTitle);

    void        SetLineStyles(char **papszStyles, int numLinesOfEachColor = 1);
    void        SetScaleType(GDC_SCALE_T eScaleType);

    char        **GetLegendSamples(const char *pszGIFSampleURL,
                                   int nWidth, int nHeight);

    const char *SaveSlideURL(const char *pszCGIPath);

    char       **GetParams();
};


#endif /* _OWTGDCHART_H_INCLUDED_ */

