/**********************************************************************
 * $Id: owtcom_htdoc.cpp,v 1.2 1999/06/09 19:27:20 daniel Exp $
 *
 * Name:     owtcom_htdoc.cpp
 * Project:  OWT HTML manipulation library - COM Interface
 * Purpose:  Implementation of the OWTCHtDoc component class
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1999, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: owtcom_htdoc.cpp,v $
 * Revision 1.2  1999/06/09 19:27:20  daniel
 * Added GetText() and GetChildText()
 *
 * Revision 1.1.1.1  1999/06/04 19:37:36  daniel
 * First kind of complete version
 *
 **********************************************************************/

#include <objbase.h>
#include <stdio.h>
#include <io.h>         // dup()

#include "Iface.h"
#include "ic_Util.h"
#include "ic_CUnknown.h"
#include "ic_CFactory.h" // Needed for module handle
#include "owtcom_cmpnt.h"
#include "oledb_sup.h"
#include "owthtcomblock.h"

#define OWT_THIS_MODULE  "OWTCHtDoc"
#define L_OWT_THIS_MODULE  L"OWTCHtDoc"  //Unicode equivalent (dbl-byte)

static inline void trace(const char* msg)
	{ Util::Trace(OWT_THIS_MODULE, msg, S_OK) ;}
static inline void trace(const char* msg, HRESULT hr)
	{ Util::Trace(OWT_THIS_MODULE, msg, hr) ;}


/*=====================================================================
 *                        (class) OWTCHtDoc
 *====================================================================*/

/**********************************************************************
 *                       OWTCHtDoc::OWTCHtDoc()
 *
 * Constructor
 **********************************************************************/
OWTCHtDoc::OWTCHtDoc(IUnknown* pUnknownOuter): 
                                          CUnknown(pUnknownOuter), 
                                          m_pITypeInfo(NULL)
{
    /* Create instance of OWTHtDoc 
     */
    m_pDoc = new OWTHtDoc;
}

/**********************************************************************
 *                       OWTCHtDoc::~OWTCHtDoc()
 *
 * Destructor
 **********************************************************************/
OWTCHtDoc::~OWTCHtDoc()
{
    trace("Destroy self.");

    if (m_pITypeInfo != NULL)
    {
        m_pITypeInfo->Release() ;
    }

    if (m_pDoc)
        delete m_pDoc;
}



/**********************************************************************
 *            NondelegatingQueryInterface implementation
 **********************************************************************/
HRESULT __stdcall OWTCHtDoc::NondelegatingQueryInterface(const IID& iid,
                                                           void** ppv)
{ 	
    if (iid == IID_IOWTCHtDoc)
    {
        return FinishQI(static_cast<IOWTCHtDoc*>(this), ppv) ;
    }
    else if (iid == IID_IDispatch)
    {
        trace("Queried for IDispatch.") ;
        return FinishQI(static_cast<IDispatch*>(this), ppv) ;
    }
    else if (iid == IID_ISupportErrorInfo)
    {
        trace("Queried for ISupportErrorInfo.") ;
        return FinishQI(static_cast<ISupportErrorInfo*>(this), ppv) ;
    }
    else
    {
        return CUnknown::NondelegatingQueryInterface(iid, ppv) ;
    }
}


/**********************************************************************
 *                       OWTCHtDoc::Init()
 *
 * Init new object instance... load and register the type library.
 * The type library is stored in the DLL.
 **********************************************************************/
HRESULT OWTCHtDoc::Init()
{
    HRESULT hr ;

    // Load TypeInfo on demand if we haven't already loaded it.
    if (m_pITypeInfo == NULL)
    {
        ITypeLib* pITypeLib = NULL ;
        hr = ::LoadRegTypeLib(LIBID_OWTCOMLib, 
                              1, 0, // Major/Minor version numbers
                              0x00, 
                              &pITypeLib) ;
        if (FAILED(hr)) 
        {
            trace("LoadRegTypeLib Failed, now trying LoadTypeLib.", hr) ;
            // If it wasn't registered, try to load it from the path.

            // Get the fullname of the server's executable...
            // the type lib is storefd in it.
            char szModule[512] ;
            DWORD dwResult = ::GetModuleFileName(CFactory::s_hModule,
                                                 szModule,
                                                 512) ; 

            // convert to wide char
            wchar_t wszTypeLibFullName[_MAX_PATH] ;
            mbstowcs(wszTypeLibFullName, szModule, _MAX_PATH) ;

            // if LoadTypeLib succeeds, it will have registered
            // the type library for us.
            // for the next time.  
            hr = ::LoadTypeLib(wszTypeLibFullName,
                               &pITypeLib) ;
            if(FAILED(hr))        
            {
                trace("LoadTypeLib Failed.", hr) ;
                return hr;   
            }

            // Ensure that the type library is registered.
            hr = RegisterTypeLib(pITypeLib, wszTypeLibFullName, NULL) ;
            if(FAILED(hr))        
            {
                trace("RegisterTypeLib Failed.", hr) ;
                return hr ;   
            }
        }

        // Get type information for the interface of the object.
        hr = pITypeLib->GetTypeInfoOfGuid(IID_IOWTCHtDoc,
                                          &m_pITypeInfo) ;
        pITypeLib->Release() ;
        if (FAILED(hr))  
        { 
            trace("GetTypeInfoOfGuid failed.", hr) ;
            return hr ;
        }   
    }
    return S_OK ;
}

/**********************************************************************
 *                       OWTCHtDoc::CreateInstance()
 *
 * Static Creation function used by CFactory
 **********************************************************************/
HRESULT OWTCHtDoc::CreateInstance(IUnknown* pUnknownOuter,
                                    CUnknown** ppNewComponent ) 
{
    if (pUnknownOuter != NULL)
    {
        // Don't allow aggregation (just for the heck of it).
        return CLASS_E_NOAGGREGATION ;
    }

    *ppNewComponent = new OWTCHtDoc(pUnknownOuter) ;
    return S_OK ;
}

///////////////////////////////////////////////////////////
//
// IDispatch implementation
//
HRESULT __stdcall OWTCHtDoc::GetTypeInfoCount(UINT* pCountTypeInfo)
{
    trace("GetTypeInfoCount call succeeded.") ;
    *pCountTypeInfo = 1 ;
    return S_OK ;
}

HRESULT __stdcall OWTCHtDoc::GetTypeInfo(
            UINT iTypeInfo,
            LCID,          // This object does not support localization.
            ITypeInfo** ppITypeInfo)
{    
    *ppITypeInfo = NULL ;
    
    if(iTypeInfo != 0)
    {
        trace("GetTypeInfo call failed -- bad iTypeInfo index.") ;
        return DISP_E_BADINDEX ; 
    }

    trace("GetTypeInfo call succeeded.") ;

    // Call AddRef and return the pointer.
    m_pITypeInfo->AddRef() ; 
    *ppITypeInfo = m_pITypeInfo ;
    return S_OK ;
}

HRESULT __stdcall OWTCHtDoc::GetIDsOfNames(  
            const IID& iid,
            OLECHAR** arrayNames,
            UINT countNames,
            LCID,          // Localization is not supported.
            DISPID* arrayDispIDs)
{
    if (iid != IID_NULL)
    {
        trace("GetIDsOfNames call failed -- bad IID.") ;
        return DISP_E_UNKNOWNINTERFACE ;
    }

    trace("GetIDsOfNames call succeeded.") ;
    HRESULT hr = m_pITypeInfo->GetIDsOfNames(arrayNames,
                                             countNames,
                                             arrayDispIDs) ;
    return hr ;
}

HRESULT __stdcall OWTCHtDoc::Invoke(   
            DISPID dispidMember,
            const IID& iid,
            LCID,          // Localization is not supported.
            WORD wFlags,
            DISPPARAMS* pDispParams,
            VARIANT* pvarResult,
            EXCEPINFO* pExcepInfo,
            UINT* pArgErr)
{        
    if (iid != IID_NULL)
    {
        trace("Invoke call failed -- bad IID.") ;
        return DISP_E_UNKNOWNINTERFACE ;
    }

    ::SetErrorInfo(0, NULL) ;

    trace("Invoke call succeeded.") ;
    HRESULT hr = m_pITypeInfo->Invoke(static_cast<IDispatch*>(this),
                                      dispidMember, wFlags, pDispParams,
                                      pvarResult, pExcepInfo, pArgErr) ; 
    return hr ;
}

/**********************************************************************
 *                 OWTCHtDoc::RaiseException()
 *
 * Helper method to facilitate handling error situations.
 **********************************************************************/
HRESULT OWTCHtDoc::RaiseException(HRESULT hrToReturn, const char *msg)
{
    HRESULT hr;
    LPOLESTR pwszMsg=NULL;
    ICreateErrorInfo *pICreateErr;

    trace("----- Raising Exception: -----");
    trace(msg);
    trace("----- -----");

    // Create the error info object.
    hr = ::CreateErrorInfo(&pICreateErr);
    if (FAILED(hr))
    {
        trace("RaiseException failed");
        return E_FAIL;
    }

    // pICreateErr->SetHelpFile(...);
    // pICreateErr->SetHelpContext(...);
    pICreateErr->SetSource(L_OWT_THIS_MODULE);
    hr = AnsiToUnicode(msg, &pwszMsg);
    if (SUCCEEDED(hr))
    {
        pICreateErr->SetDescription(pwszMsg);
        CoTaskMemFree(pwszMsg);
        pwszMsg = NULL;
    }

    IErrorInfo *pIErrorInfo = NULL;
    hr = pICreateErr->QueryInterface(IID_IErrorInfo, (void**)&pIErrorInfo);
    if (SUCCEEDED(hr))
    {
        ::SetErrorInfo(0L, pIErrorInfo);
        pIErrorInfo->Release();
    }
    pICreateErr->Release();

    return hrToReturn;
}


///////////////////////////////////////////////////////////////////////
//           OWTCHtDoc::   IOWTCHtDoc interface implementation
///////////////////////////////////////////////////////////////////////

/**********************************************************************
 *                 OWTCHtDoc::Draw()
 *
 * vFileHandle is optional, if it is not present then stdout will be used.
 * otherwise the corresponding file is opened using fdopen()
 **********************************************************************/
HRESULT __stdcall OWTCHtDoc::Draw(VARIANT vFileHandle )
{ 
    FILE *fpOut = NULL;
    int  nFileHandle;

    trace("Draw()");

    /*-------------------------------------------------------------
     * Extract file no from the optional VARIANT argument
     *------------------------------------------------------------*/
    nFileHandle = -1;
    if (vFileHandle.vt != VT_ERROR)
    {
        HRESULT hr;
        VARIANTARG  varHandle;
        VariantInit(&varHandle);
        hr = VariantChangeType(&varHandle, &vFileHandle, 0, VT_I4);
        if (FAILED(hr))
        {
            return RaiseException(E_INVALIDARG, "Draw(): Invalid argument.");
        }
        nFileHandle = varHandle.lVal;
    }

    char szBuf[100];
    sprintf(szBuf, "Output file handle is %d", nFileHandle);
    trace(szBuf);

    if (nFileHandle >= 0)
    {
        /*-------------------------------------------------------------
         * dup() the file handle, and get a stream handle from it.
         * We need to dup() it in order to be allowed to call fclose()
         * to release the stream handle when we're done.
         *------------------------------------------------------------*/
        nFileHandle = dup(nFileHandle);
        if (nFileHandle > 0)
            fpOut = fdopen(nFileHandle, "a");
    }
    else
    {
        fpOut = stdout;
    }

    if (fpOut == NULL)
    {
        return RaiseException(E_ACCESSDENIED, 
                              "Draw() cannot access specified file handle.");
    }

    m_pDoc->Draw(fpOut);
    fflush(fpOut);

    /*-------------------------------------------------------------
     *Close file, unless we're writing to stdout
     *------------------------------------------------------------*/
    if (nFileHandle >= 0)
        fclose(fpOut);

    return S_OK ;
}

/**********************************************************************
 *                 OWTCHtDoc::SetText()
 *
 * Valid input types are BSTR or a SAFEARRAY of BSTR.
 **********************************************************************/
HRESULT __stdcall OWTCHtDoc::SetText( VARIANT vText )
{ 
    HRESULT     hr;
    char        **papszText = NULL;

    trace("SetText()") ;

    // The magic of extracting the data from the VARIANT is hidden
    // inside BSTRArrayToStringlist()

    hr = BSTRArrayToStringlist(vText, &papszText);
    if (FAILED(hr))
        return hr;

    // Pass the text value to the OWTHtBlock... it will be responsible
    // to eventually destroy the stringlist.

    m_pDoc->SetText(papszText, FALSE);

    return S_OK ;
}

/**********************************************************************
 *                 OWTCHtDoc::SetChildText()
 *
 * Valid input types are BSTR or a SAFEARRAY of BSTR.
 **********************************************************************/
HRESULT __stdcall OWTCHtDoc::SetChildText(BSTR bstrName, VARIANT vText )
{ 
    HRESULT     hr;
    char        *pszName=NULL, **papszText = NULL;

    trace("SetChildText()") ;

    // The magic of extracting the data from the VARIANT is hidden
    // inside BSTRArrayToStringlist()

    hr = BSTRArrayToStringlist(vText, &papszText);
    if (FAILED(hr))
        return hr;

    // Pass the text value to the OWTHtBlock... it will be responsible
    // to eventually destroy the stringlist.

    hr = UnicodeToAnsi( bstrName, &pszName  );
    if( FAILED(hr) )
        return hr;
    m_pDoc->SetChildText(pszName, papszText, FALSE);
    CoTaskMemFree(pszName);

    return S_OK ;
}

/**********************************************************************
 *                 OWTCHtDoc::GetName()
 *
 **********************************************************************/
HRESULT __stdcall OWTCHtDoc::GetName(BSTR *pbstrNameOut )
{ 
    const char  *pszName = NULL;
    HRESULT     hr;

    trace("GetName()") ;

    return AnsiToBSTR(m_pDoc->GetName(), pbstrNameOut);
}

/**********************************************************************
 *                 OWTCHtDoc::SetName()
 *
 **********************************************************************/
HRESULT __stdcall OWTCHtDoc::SetName(BSTR bstrName )
{ 
    char        *pszName = NULL;
    HRESULT     hr;

    trace("SetName()") ;

    hr = UnicodeToAnsi( bstrName, &pszName  );
    if( FAILED(hr) )
        return hr;
    m_pDoc->SetName(pszName);
    CoTaskMemFree(pszName);

    return S_OK ;
}

/**********************************************************************
 *                 OWTCHtDoc::ChildExists()
 *
 * Test if a child with the given name exists.  
 * Returns VARIANT_TRUE if the child exists, or VARIANT_FALSE otherwise.
 **********************************************************************/
HRESULT __stdcall OWTCHtDoc::ChildExists(BSTR bstrName, 
                                         VARIANT_BOOL *pvbRetVal )
{ 
    char        *pszName = NULL;
    HRESULT     hr;

    trace("ChildExists()") ;

    hr = UnicodeToAnsi( bstrName, &pszName  );
    if( FAILED(hr) )
        return hr;
    if ( m_pDoc->GetChild(pszName) != NULL)
        *pvbRetVal = VARIANT_TRUE;
    else
        *pvbRetVal = VARIANT_FALSE;

    CoTaskMemFree(pszName);

    return S_OK ;
}

/**********************************************************************
 *                 OWTCHtDoc::SetChild()
 *
 * Pass a COM component as a child of this document.  The component 
 * will be encapsulated inside a OWTHtComBlock object before it is
 * stored in the actual document.
 **********************************************************************/
HRESULT __stdcall OWTCHtDoc::SetChild(BSTR bstrName, 
                                      IUnknown *pIUnknown)
{ 
    char        *pszName = NULL;
    HRESULT     hr = S_OK;
    OWTHtComBlock *pComBlock = NULL;

    trace("SetChild()") ;

    /*-----------------------------------------------------------------
     * First make sure a child with that name exists
     *----------------------------------------------------------------*/
    hr = UnicodeToAnsi( bstrName, &pszName  );
    if( FAILED(hr) )
        return hr;

    hr = S_OK;
    if ( m_pDoc->GetChild(pszName) != NULL)
    {
        /*-------------------------------------------------------------
         * Create a wrapper object and pass it to the document
         * If anytihng fails then OWTHtComBlock(() calls CPLError()
         *------------------------------------------------------------*/
        pComBlock = new OWTHtComBlock(pIUnknown);

        if (CPLGetLastErrorNo() == 0)
        {
            m_pDoc->SetChild(pszName, pComBlock);
        }
        else
        {
            delete pComBlock;
            hr = RaiseException(E_FAIL, CPLGetLastErrorMsg());
        }
    }
    else
    {
        hr = RaiseException(E_FAIL, "SetChild() - Child not found");
    }

    CoTaskMemFree(pszName);

    return hr ;
}

/**********************************************************************
 *                 OWTCHtDoc::GetText()
 *
 * Return the static text for this object in a SAFEARRAY of BSTR.
 *
 * Note that this function will return an empty array if it is called
 * after the object has been parsed.
 **********************************************************************/
HRESULT __stdcall OWTCHtDoc::GetText(VARIANT *pvabstrTextOut )
{ 
    HRESULT     hr;

    trace("GetText()") ;

    return StringlistToBSTRArray(m_pDoc->GetText(), pvabstrTextOut);
}

/**********************************************************************
 *                 OWTCHtDoc::GetChildText()
 *
 * Return the static text for a child of this object in a SAFEARRAY of BSTR.
 *
 * Note that this function will return an empty array if it is called
 * after the object has been parsed.
 *
 * An error will be produced if the named children does not exist.
 **********************************************************************/
HRESULT __stdcall OWTCHtDoc::GetChildText(BSTR bstrName,
                                            VARIANT *pvabstrTextOut )
{ 
    char        **papszText = NULL;
    char        *pszName = NULL;
    HRESULT     hr;

    trace("GetChildText()") ;

    hr = UnicodeToAnsi( bstrName, &pszName  );
    if( FAILED(hr) )
        return hr;
    papszText = m_pDoc->GetChildText(pszName);

    if (CPLGetLastErrorNo() != 0)
    {   
        // Likely that child does not exist.
        hr = RaiseException(E_FAIL, CPLGetLastErrorMsg());
    }
    else
    {
        hr = StringlistToBSTRArray(papszText, pvabstrTextOut);
    }

    CoTaskMemFree(pszName);

    return hr;
}



/**********************************************************************
 *                 OWTCHtDoc::IOWTCHtDoc interface implementation
 **********************************************************************/

/**********************************************************************
 *                 OWTCHtDoc::LoadFromFile()
 *
 **********************************************************************/
HRESULT __stdcall OWTCHtDoc::LoadFromFile(BSTR bstrFilename )
{ 
    char        *pszFname = NULL;
    HRESULT     hr = S_OK;
    GBool       bStatus;

    trace("LoadFromFile()") ;

    hr = UnicodeToAnsi( bstrFilename, &pszFname  );
    if( FAILED(hr) )
        return hr;
    bStatus = m_pDoc->LoadFromFile(pszFname);
    CoTaskMemFree(pszFname);

    if (bStatus == FALSE)
    {
        hr = HRESULT_FROM_WIN32(ERROR_FILE_NOT_FOUND);
        return RaiseException(E_FAIL, CPLGetLastErrorMsg());
    }

    return S_OK ;
}

/**********************************************************************
 *                 OWTCHtDoc::SaveToFile()
 *
 **********************************************************************/
HRESULT __stdcall OWTCHtDoc::SaveToFile(BSTR bstrFilename )
{ 
    char        *pszFname = NULL;
    HRESULT     hr = S_OK;
    GBool       bStatus;

    trace("SaveToFile()") ;

    hr = UnicodeToAnsi( bstrFilename, &pszFname  );
    if( FAILED(hr) )
        return hr;
    bStatus = m_pDoc->SaveToFile(pszFname);
    CoTaskMemFree(pszFname);

    if (bStatus == FALSE)
    {
        hr = HRESULT_FROM_WIN32(ERROR_ACCESS_DENIED);
        return RaiseException(E_FAIL, CPLGetLastErrorMsg());
    }

    return S_OK ;
}

/**********************************************************************
 *                 OWTCHtDoc::SetContentType()
 *
 **********************************************************************/
HRESULT __stdcall OWTCHtDoc::SetContentType(BSTR bstrContentType )
{ 
    char        *pszContentType = NULL;
    HRESULT     hr;

    trace("SetContentType()") ;

    hr = UnicodeToAnsi( bstrContentType, &pszContentType  );
    if( FAILED(hr) )
        return hr;
    m_pDoc->SetContentType(pszContentType);
    CoTaskMemFree(pszContentType);

    return S_OK ;
}
