/**********************************************************************
 * $Id: owtcom_htoutput.cpp,v 1.1.1.1 1999/06/04 19:37:36 daniel Exp $
 *
 * Name:     owtcom_htoutput.cpp
 * Project:  OWT HTML manipulation library - COM Interface
 * Purpose:  Implemetation of the OWTCHtOutput component class
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1999, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: owtcom_htoutput.cpp,v $
 * Revision 1.1.1.1  1999/06/04 19:37:36  daniel
 * First kind of complete version
 *
 **********************************************************************/

#include <objbase.h>
#include <stdio.h>
#include <io.h>         // dup()

#include "Iface.h"
#include "ic_Util.h"
#include "ic_CUnknown.h"
#include "ic_CFactory.h" // Needed for module handle
#include "owtcom_cmpnt.h"
#include "oledb_sup.h"

#define OWT_THIS_MODULE  "OWTCHtOutput"
#define L_OWT_THIS_MODULE  L"OWTCHtOutput"  //Unicode equivalent (dbl-byte)

static inline void trace(const char* msg)
	{ Util::Trace(OWT_THIS_MODULE, msg, S_OK) ;}
static inline void trace(const char* msg, HRESULT hr)
	{ Util::Trace(OWT_THIS_MODULE, msg, hr) ;}

/*=====================================================================
 *                        (class) OWTCHtOutput
 *====================================================================*/

/**********************************************************************
 *                       OWTCHtOutput::OWTCHtOutput()
 *
 * Constructor
 **********************************************************************/
OWTCHtOutput::OWTCHtOutput(IUnknown* pUnknownOuter): 
                                          CUnknown(pUnknownOuter), 
                                          m_pITypeInfo(NULL)

{
    /* Default output file is stdout when m_fpOut = NULL
     */
    m_fpOut = NULL;
}

/**********************************************************************
 *                       OWTCHtOutput::~OWTCHtOutput()
 *
 * Destructor
 **********************************************************************/
OWTCHtOutput::~OWTCHtOutput()
{
    trace("OWTCHtOutput: Destroy self.");

    if (m_pITypeInfo != NULL)
    {
        m_pITypeInfo->Release() ;
    }

    /*-------------------------------------------------------------
     *Close file, unless we're writing to stdout
     *------------------------------------------------------------*/
    if (m_fpOut != NULL)
        fclose(m_fpOut);
}

/**********************************************************************
 *            NondelegatingQueryInterface implementation
 **********************************************************************/
HRESULT __stdcall OWTCHtOutput::NondelegatingQueryInterface(const IID& iid,
                                                           void** ppv)
{ 	
    if (iid == IID_IOWTCHtOutput)
    {
        return FinishQI(static_cast<IOWTCHtOutput*>(this), ppv) ;
    }
    else if (iid == IID_IDispatch)
    {
        trace("Queried for IDispatch.") ;
        return FinishQI(static_cast<IDispatch*>(this), ppv) ;
    }
    else if (iid == IID_ISupportErrorInfo)
    {
        trace("Queried for ISupportErrorInfo.") ;
        return FinishQI(static_cast<ISupportErrorInfo*>(this), ppv) ;
    }
    else
    {
        return CUnknown::NondelegatingQueryInterface(iid, ppv) ;
    }
}


/**********************************************************************
 *                       OWTCHtOutput::Init()
 *
 * Init new object instance... load and register the type library.
 * The type library is stored in the DLL.
 **********************************************************************/
HRESULT OWTCHtOutput::Init()
{
    HRESULT hr ;

    // Load TypeInfo on demand if we haven't already loaded it.
    if (m_pITypeInfo == NULL)
    {
        ITypeLib* pITypeLib = NULL ;
        hr = ::LoadRegTypeLib(LIBID_OWTCOMLib, 
                              1, 0, // Major/Minor version numbers
                              0x00, 
                              &pITypeLib) ;
        if (FAILED(hr)) 
        {
            trace("LoadRegTypeLib Failed, now trying LoadTypeLib.", hr) ;
            // If it wasn't registered, try to load it from the path.

            // Get the fullname of the server's executable...
            // the type lib is storefd in it.
            char szModule[512] ;
            DWORD dwResult = ::GetModuleFileName(CFactory::s_hModule,
                                                 szModule,
                                                 512) ; 

            // convert to wide char
            wchar_t wszTypeLibFullName[_MAX_PATH] ;
            mbstowcs(wszTypeLibFullName, szModule, _MAX_PATH) ;

            // if LoadTypeLib succeeds, it will have registered
            // the type library for us.
            // for the next time.  
            hr = ::LoadTypeLib(wszTypeLibFullName,
                               &pITypeLib) ;
            if(FAILED(hr))        
            {
                trace("LoadTypeLib Failed.", hr) ;
                return hr;   
            }

            // Ensure that the type library is registered.
            hr = RegisterTypeLib(pITypeLib, wszTypeLibFullName, NULL) ;
            if(FAILED(hr))        
            {
                trace("RegisterTypeLib Failed.", hr) ;
                return hr ;   
            }
        }

        // Get type information for the interface of the object.
        hr = pITypeLib->GetTypeInfoOfGuid(IID_IOWTCHtOutput,
                                          &m_pITypeInfo) ;
        pITypeLib->Release() ;
        if (FAILED(hr))  
        { 
            trace("GetTypeInfoOfGuid failed.", hr) ;
            return hr ;
        }   
    }
    return S_OK ;
}

/**********************************************************************
 *                       OWTCHtOutput::CreateInstance()
 *
 * Static Creation function used by CFactory
 **********************************************************************/
HRESULT OWTCHtOutput::CreateInstance(IUnknown* pUnknownOuter,
                                    CUnknown** ppNewComponent ) 
{
    if (pUnknownOuter != NULL)
    {
        // Don't allow aggregation (just for the heck of it).
        return CLASS_E_NOAGGREGATION ;
    }

    *ppNewComponent = new OWTCHtOutput(pUnknownOuter) ;
    return S_OK ;
}

///////////////////////////////////////////////////////////
//
// IDispatch implementation
//
HRESULT __stdcall OWTCHtOutput::GetTypeInfoCount(UINT* pCountTypeInfo)
{
    trace("GetTypeInfoCount call succeeded.") ;
    *pCountTypeInfo = 1 ;
    return S_OK ;
}

HRESULT __stdcall OWTCHtOutput::GetTypeInfo(
            UINT iTypeInfo,
            LCID,          // This object does not support localization.
            ITypeInfo** ppITypeInfo)
{    
    *ppITypeInfo = NULL ;
    
    if(iTypeInfo != 0)
    {
        trace("GetTypeInfo call failed -- bad iTypeInfo index.") ;
        return DISP_E_BADINDEX ; 
    }

    trace("GetTypeInfo call succeeded.") ;

    // Call AddRef and return the pointer.
    m_pITypeInfo->AddRef() ; 
    *ppITypeInfo = m_pITypeInfo ;
    return S_OK ;
}

HRESULT __stdcall OWTCHtOutput::GetIDsOfNames(  
            const IID& iid,
            OLECHAR** arrayNames,
            UINT countNames,
            LCID,          // Localization is not supported.
            DISPID* arrayDispIDs)
{
    if (iid != IID_NULL)
    {
        trace("GetIDsOfNames call failed -- bad IID.") ;
        return DISP_E_UNKNOWNINTERFACE ;
    }

    trace("GetIDsOfNames call succeeded.") ;
    HRESULT hr = m_pITypeInfo->GetIDsOfNames(arrayNames,
                                             countNames,
                                             arrayDispIDs) ;
    return hr ;
}

HRESULT __stdcall OWTCHtOutput::Invoke(   
            DISPID dispidMember,
            const IID& iid,
            LCID,          // Localization is not supported.
            WORD wFlags,
            DISPPARAMS* pDispParams,
            VARIANT* pvarResult,
            EXCEPINFO* pExcepInfo,
            UINT* pArgErr)
{        
    if (iid != IID_NULL)
    {
        trace("Invoke call failed -- bad IID.") ;
        return DISP_E_UNKNOWNINTERFACE ;
    }

    ::SetErrorInfo(0, NULL) ;

    trace("Invoke call succeeded.") ;
    HRESULT hr = m_pITypeInfo->Invoke(static_cast<IDispatch*>(this),
                                      dispidMember, wFlags, pDispParams,
                                      pvarResult, pExcepInfo, pArgErr) ; 
    return hr ;
}


/**********************************************************************
 *                 OWTCHtOutput::RaiseException()
 *
 * Helper method to facilitate handling error situations.
 **********************************************************************/
HRESULT OWTCHtOutput::RaiseException(HRESULT hrToReturn, const char *msg)
{
    HRESULT hr;
    LPOLESTR pwszMsg=NULL;
    ICreateErrorInfo *pICreateErr;

    trace("----- Raising Exception: -----");
    trace(msg);
    trace("----- -----");

    // Create the error info object.
    hr = ::CreateErrorInfo(&pICreateErr);
    if (FAILED(hr))
    {
        trace("RaiseException failed");
        return E_FAIL;
    }

    // pICreateErr->SetHelpFile(...);
    // pICreateErr->SetHelpContext(...);
    pICreateErr->SetSource(L_OWT_THIS_MODULE);
    hr = AnsiToUnicode(msg, &pwszMsg);
    if (SUCCEEDED(hr))
    {
        pICreateErr->SetDescription(pwszMsg);
        CoTaskMemFree(pwszMsg);
        pwszMsg = NULL;
    }

    IErrorInfo *pIErrorInfo = NULL;
    hr = pICreateErr->QueryInterface(IID_IErrorInfo, (void**)&pIErrorInfo);
    if (SUCCEEDED(hr))
    {
        ::SetErrorInfo(0L, pIErrorInfo);
        pIErrorInfo->Release();
    }
    pICreateErr->Release();

    return hrToReturn;
}


///////////////////////////////////////////////////////////////////////
//           OWTCHtOutput::   IOWTCHtOutput interface implementation
///////////////////////////////////////////////////////////////////////

/**********************************************************************
 *                 OWTCHtOutput::SetFileHandle()
 *
 * We will access nfileHandle using fdopen()
 *
 * Passing -1 will use stdout.
 **********************************************************************/
HRESULT __stdcall OWTCHtOutput::SetFileHandle(LONG nFileHandle )
{ 
    FILE *fpOut = NULL;

    trace("SetFileHandle()");

    /*-------------------------------------------------------------
     * Close current output file if any.
     *------------------------------------------------------------*/
    if (m_fpOut)
    {
        fclose(m_fpOut);
        m_fpOut = NULL;
    }

    char szBuf[100];
    sprintf(szBuf, "OWTHtOutput file handle is %d", nFileHandle);
    trace(szBuf);

    if (nFileHandle >= 0)
    {
        /*-------------------------------------------------------------
         * dup() the file handle, and get a stream handle from it.
         * We need to dup() it in order to be allowed to call fclose()
         * to release the stream handle when we're done.
         *------------------------------------------------------------*/
        nFileHandle = dup(nFileHandle);
        if (nFileHandle > 0)
            m_fpOut = fdopen(nFileHandle, "a");

        if (m_fpOut == NULL)
        {
            return RaiseException(E_FAIL,
                                  "Cannot access specified file handle.");
        }
    }
    else
    {
        // Use stdout
        m_fpOut = NULL;
    }

    return S_OK ;
}

/**********************************************************************
 *                 OWTCHtOutput::Write()
 *
 * Write some text to the output device.
 *
 * Valid input types are BSTR or a SAFEARRAY of BSTR.
 **********************************************************************/
HRESULT __stdcall OWTCHtOutput::Write( VARIANT vText )
{ 
    HRESULT     hr;
    char        **papszText = NULL;

    trace("Write()") ;

    // The magic of extracting the data from the VARIANT is hidden
    // inside BSTRArrayToStringlist()

    hr = BSTRArrayToStringlist(vText, &papszText);
    if (FAILED(hr))
        return hr;

    // Write the text to the file and destroy it.

    if (m_fpOut == NULL)
    {
        CSLPrint(papszText, stdout);
        fflush(stdout);
    }
    else
    {
        CSLPrint(papszText, m_fpOut);
        fflush(m_fpOut);
    }

    CSLDestroy(papszText);

    return S_OK ;
}
