/**********************************************************************
 * $Id: owtcom_httable.cpp,v 1.1.1.1 1999/06/04 19:37:36 daniel Exp $
 *
 * Name:     owtcom_httable.cpp
 * Project:  OWT HTML manipulation library - COM Interface
 * Purpose:  Implementation of the OWTCHtTable component class
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1999, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: owtcom_httable.cpp,v $
 * Revision 1.1.1.1  1999/06/04 19:37:36  daniel
 * First kind of complete version
 *
 **********************************************************************/

#include <objbase.h>
#include <stdio.h>
#include <io.h>         // dup()

#include "Iface.h"
#include "ic_Util.h"
#include "ic_CUnknown.h"
#include "ic_CFactory.h" // Needed for module handle
#include "owtcom_cmpnt.h"
#include "oledb_sup.h"
#include "owthtcomblock.h"

#define OWT_THIS_MODULE  "OWTCHtTable"
#define L_OWT_THIS_MODULE  L"OWTCHtTable"  //Unicode equivalent (dbl-byte)

static inline void trace(const char* msg)
	{ Util::Trace(OWT_THIS_MODULE, msg, S_OK) ;}
static inline void trace(const char* msg, HRESULT hr)
	{ Util::Trace(OWT_THIS_MODULE, msg, hr) ;}


/*=====================================================================
 *                        (class) OWTCHtTable
 *====================================================================*/

/**********************************************************************
 *                       OWTCHtTable::OWTCHtTable()
 *
 * Constructor
 **********************************************************************/
OWTCHtTable::OWTCHtTable(IUnknown* pUnknownOuter): 
                                          CUnknown(pUnknownOuter), 
                                          m_pITypeInfo(NULL)
{
    /* Create instance of OWTHtTable 
     */
    m_pTable = new OWTHtTable;
}

/**********************************************************************
 *                       OWTCHtTable::~OWTCHtTable()
 *
 * Destructor
 **********************************************************************/
OWTCHtTable::~OWTCHtTable()
{
    trace("Destroy self.");

    if (m_pITypeInfo != NULL)
    {
        m_pITypeInfo->Release() ;
    }

    if (m_pTable)
        delete m_pTable;
}



/**********************************************************************
 *            NondelegatingQueryInterface implementation
 **********************************************************************/
HRESULT __stdcall OWTCHtTable::NondelegatingQueryInterface(const IID& iid,
                                                           void** ppv)
{ 	
    if (iid == IID_IOWTCHtTable)
    {
        return FinishQI(static_cast<IOWTCHtTable*>(this), ppv) ;
    }
    else if (iid == IID_IDispatch)
    {
        trace("Queried for IDispatch.") ;
        return FinishQI(static_cast<IDispatch*>(this), ppv) ;
    }
    else if (iid == IID_ISupportErrorInfo)
    {
        trace("Queried for ISupportErrorInfo.") ;
        return FinishQI(static_cast<ISupportErrorInfo*>(this), ppv) ;
    }
    else
    {
        return CUnknown::NondelegatingQueryInterface(iid, ppv) ;
    }
}


/**********************************************************************
 *                       OWTCHtTable::Init()
 *
 * Init new object instance... load and register the type library.
 * The type library is stored in the DLL.
 **********************************************************************/
HRESULT OWTCHtTable::Init()
{
    HRESULT hr ;

    // Load TypeInfo on demand if we haven't already loaded it.
    if (m_pITypeInfo == NULL)
    {
        ITypeLib* pITypeLib = NULL ;
        hr = ::LoadRegTypeLib(LIBID_OWTCOMLib, 
                              1, 0, // Major/Minor version numbers
                              0x00, 
                              &pITypeLib) ;
        if (FAILED(hr)) 
        {
            trace("LoadRegTypeLib Failed, now trying LoadTypeLib.", hr) ;
            // If it wasn't registered, try to load it from the path.

            // Get the fullname of the server's executable...
            // the type lib is storefd in it.
            char szModule[512] ;
            DWORD dwResult = ::GetModuleFileName(CFactory::s_hModule,
                                                 szModule,
                                                 512) ; 

            // convert to wide char
            wchar_t wszTypeLibFullName[_MAX_PATH] ;
            mbstowcs(wszTypeLibFullName, szModule, _MAX_PATH) ;

            // if LoadTypeLib succeeds, it will have registered
            // the type library for us.
            // for the next time.  
            hr = ::LoadTypeLib(wszTypeLibFullName,
                               &pITypeLib) ;
            if(FAILED(hr))        
            {
                trace("LoadTypeLib Failed.", hr) ;
                return hr;   
            }

            // Ensure that the type library is registered.
            hr = RegisterTypeLib(pITypeLib, wszTypeLibFullName, NULL) ;
            if(FAILED(hr))        
            {
                trace("RegisterTypeLib Failed.", hr) ;
                return hr ;   
            }
        }

        // Get type information for the interface of the object.
        hr = pITypeLib->GetTypeInfoOfGuid(IID_IOWTCHtTable,
                                          &m_pITypeInfo) ;
        pITypeLib->Release() ;
        if (FAILED(hr))  
        { 
            trace("GetTypeInfoOfGuid failed.", hr) ;
            return hr ;
        }   
    }
    return S_OK ;
}

/**********************************************************************
 *                       OWTCHtTable::CreateInstance()
 *
 * Static Creation function used by CFactory
 **********************************************************************/
HRESULT OWTCHtTable::CreateInstance(IUnknown* pUnknownOuter,
                                    CUnknown** ppNewComponent ) 
{
    if (pUnknownOuter != NULL)
    {
        // Don't allow aggregation (just for the heck of it).
        return CLASS_E_NOAGGREGATION ;
    }

    *ppNewComponent = new OWTCHtTable(pUnknownOuter) ;
    return S_OK ;
}

///////////////////////////////////////////////////////////
//
// IDispatch implementation
//
HRESULT __stdcall OWTCHtTable::GetTypeInfoCount(UINT* pCountTypeInfo)
{
    trace("GetTypeInfoCount call succeeded.") ;
    *pCountTypeInfo = 1 ;
    return S_OK ;
}

HRESULT __stdcall OWTCHtTable::GetTypeInfo(
            UINT iTypeInfo,
            LCID,          // This object does not support localization.
            ITypeInfo** ppITypeInfo)
{    
    *ppITypeInfo = NULL ;
    
    if(iTypeInfo != 0)
    {
        trace("GetTypeInfo call failed -- bad iTypeInfo index.") ;
        return DISP_E_BADINDEX ; 
    }

    trace("GetTypeInfo call succeeded.") ;

    // Call AddRef and return the pointer.
    m_pITypeInfo->AddRef() ; 
    *ppITypeInfo = m_pITypeInfo ;
    return S_OK ;
}

HRESULT __stdcall OWTCHtTable::GetIDsOfNames(  
            const IID& iid,
            OLECHAR** arrayNames,
            UINT countNames,
            LCID,          // Localization is not supported.
            DISPID* arrayDispIDs)
{
    if (iid != IID_NULL)
    {
        trace("GetIDsOfNames call failed -- bad IID.") ;
        return DISP_E_UNKNOWNINTERFACE ;
    }

    trace("GetIDsOfNames call succeeded.") ;
    HRESULT hr = m_pITypeInfo->GetIDsOfNames(arrayNames,
                                             countNames,
                                             arrayDispIDs) ;
    return hr ;
}

HRESULT __stdcall OWTCHtTable::Invoke(   
            DISPID dispidMember,
            const IID& iid,
            LCID,          // Localization is not supported.
            WORD wFlags,
            DISPPARAMS* pDispParams,
            VARIANT* pvarResult,
            EXCEPINFO* pExcepInfo,
            UINT* pArgErr)
{        
    if (iid != IID_NULL)
    {
        trace("Invoke call failed -- bad IID.") ;
        return DISP_E_UNKNOWNINTERFACE ;
    }

    ::SetErrorInfo(0, NULL) ;

    trace("Invoke call succeeded.") ;
    HRESULT hr = m_pITypeInfo->Invoke(static_cast<IDispatch*>(this),
                                      dispidMember, wFlags, pDispParams,
                                      pvarResult, pExcepInfo, pArgErr) ; 
    return hr ;
}

/**********************************************************************
 *                 OWTCHtTable::RaiseException()
 *
 * Helper method to facilitate handling error situations.
 **********************************************************************/
HRESULT OWTCHtTable::RaiseException(HRESULT hrToReturn, const char *msg)
{
    HRESULT hr;
    LPOLESTR pwszMsg=NULL;
    ICreateErrorInfo *pICreateErr;

    trace("----- Raising Exception: -----");
    trace(msg);
    trace("----- -----");

    // Create the error info object.
    hr = ::CreateErrorInfo(&pICreateErr);
    if (FAILED(hr))
    {
        trace("RaiseException failed");
        return E_FAIL;
    }

    // pICreateErr->SetHelpFile(...);
    // pICreateErr->SetHelpContext(...);
    pICreateErr->SetSource(L_OWT_THIS_MODULE);
    hr = AnsiToUnicode(msg, &pwszMsg);
    if (SUCCEEDED(hr))
    {
        pICreateErr->SetDescription(pwszMsg);
        CoTaskMemFree(pwszMsg);
        pwszMsg = NULL;
    }

    IErrorInfo *pIErrorInfo = NULL;
    hr = pICreateErr->QueryInterface(IID_IErrorInfo, (void**)&pIErrorInfo);
    if (SUCCEEDED(hr))
    {
        ::SetErrorInfo(0L, pIErrorInfo);
        pIErrorInfo->Release();
    }
    pICreateErr->Release();

    return hrToReturn;
}


///////////////////////////////////////////////////////////////////////
//           OWTCHtTable::   IOWTCHtBlock interface implementation
///////////////////////////////////////////////////////////////////////

/**********************************************************************
 *                 OWTCHtTable::Draw()
 *
 * vFileHandle is optional, if it is not present then stdout will be used.
 * otherwise the corresponding file is opened using fdopen()
 **********************************************************************/
HRESULT __stdcall OWTCHtTable::Draw(VARIANT vFileHandle )
{ 
    FILE *fpOut = NULL;
    int  nFileHandle;

    trace("Draw()");

    /*-------------------------------------------------------------
     * Extract file no from the optional VARIANT argument
     *------------------------------------------------------------*/
    nFileHandle = -1;
    if (vFileHandle.vt != VT_ERROR)
    {
        HRESULT hr;
        VARIANTARG  varHandle;
        VariantInit(&varHandle);
        hr = VariantChangeType(&varHandle, &vFileHandle, 0, VT_I4);
        if (FAILED(hr))
        {
            return RaiseException(E_INVALIDARG,
                         "Draw(): invalid argument... expecting file handle");
        }
        nFileHandle = varHandle.lVal;
    }

    char szBuf[100];
    sprintf(szBuf, "Output file handle is %d", nFileHandle);
    trace(szBuf);

    if (nFileHandle >= 0)
    {
        /*-------------------------------------------------------------
         * dup() the file handle, and get a stream handle from it.
         * We need to dup() it in order to be allowed to call fclose()
         * to release the stream handle when we're done.
         *------------------------------------------------------------*/
        nFileHandle = dup(nFileHandle);
        if (nFileHandle > 0)
            fpOut = fdopen(nFileHandle, "a");
    }
    else
    {
        fpOut = stdout;
    }

    if (fpOut == NULL)
    {
        return RaiseException(E_ACCESSDENIED,
                              "Draw() cannot access specified file handle.");
    }

    m_pTable->Draw(fpOut);
    fflush(fpOut);

    /*-------------------------------------------------------------
     *Close file, unless we're writing to stdout
     *------------------------------------------------------------*/
    if (nFileHandle >= 0)
        fclose(fpOut);

    return S_OK ;
}

/**********************************************************************
 *                 OWTCHtTable::SetText()
 *
 * Valid input types are BSTR or a SAFEARRAY of BSTR.
 **********************************************************************/
HRESULT __stdcall OWTCHtTable::SetText( VARIANT vText )
{ 
    HRESULT     hr;
    char        **papszText = NULL;

    trace("SetText()") ;

    // The magic of extracting the data from the VARIANT is hidden
    // inside BSTRArrayToStringlist()

    hr = BSTRArrayToStringlist(vText, &papszText);
    if (FAILED(hr))
        return hr;

    // Pass the text value to the OWTHtBlock... it will be responsible
    // to eventually destroy the stringlist.

    m_pTable->SetText(papszText, FALSE);

    return S_OK ;
}


/**********************************************************************
 *                 OWTCHtTable::GetName()
 *
 **********************************************************************/
HRESULT __stdcall OWTCHtTable::GetName(BSTR *pbstrNameOut )
{ 
    const char  *pszName = NULL;
    HRESULT     hr;

    trace("GetName()") ;

    return AnsiToBSTR(m_pTable->GetName(), pbstrNameOut);
}

/**********************************************************************
 *                 OWTCHtTable::SetName()
 *
 **********************************************************************/
HRESULT __stdcall OWTCHtTable::SetName(BSTR bstrName )
{ 
    char        *pszName = NULL;
    HRESULT     hr;

    trace("SetName()") ;

    hr = UnicodeToAnsi( bstrName, &pszName  );
    if( FAILED(hr) )
        return hr;
    m_pTable->SetName(pszName);
    CoTaskMemFree(pszName);

    return S_OK ;
}


/**********************************************************************
 *                 OWTCHtTable::IOWTCHtTable interface implementation
 **********************************************************************/

/**********************************************************************
 *                 OWTCHtTable::AddColumn()
 *
 * Valid input types are BSTR or a SAFEARRAY of BSTR.
 * vabstrData is optional... if it si not passed then the cells for the
 * new column are left empty.
 **********************************************************************/
HRESULT __stdcall OWTCHtTable::AddColumn( VARIANT vabstrData )
{ 
    HRESULT     hr;
    char        **papszCells = NULL;

    trace("AddColumn()") ;

    if (vabstrData.vt != VT_ERROR)
    {
        // The magic of extracting the data from the VARIANT is hidden
        // inside BSTRArrayToStringlist()
        hr = BSTRArrayToStringlist(vabstrData, &papszCells);
        if (FAILED(hr))
            return hr;
    }

    // Pass the cell values to the OWTHtTable and destroy the stringlist.

    if (m_pTable->AddColumn(papszCells) == 0)
        hr = S_OK;
    else
    {
        hr = RaiseException(E_FAIL, CPLGetLastErrorMsg());
    }
    CSLDestroy(papszCells);

    return hr;
}

/**********************************************************************
 *                 OWTCHtTable::AddRow()
 *
 * Valid input types are BSTR or a SAFEARRAY of BSTR.
 * vabstrData is optional... if it si not passed then the cells for the
 * new Row are left empty.
 **********************************************************************/
HRESULT __stdcall OWTCHtTable::AddRow( VARIANT vabstrData )
{ 
    HRESULT     hr;
    char        **papszCells = NULL;

    trace("AddRow()") ;

    if (vabstrData.vt != VT_ERROR)
    {
        // The magic of extracting the data from the VARIANT is hidden
        // inside BSTRArrayToStringlist()
        hr = BSTRArrayToStringlist(vabstrData, &papszCells);
        if (FAILED(hr))
            return hr;
    }

    // Pass the cell values to the OWTHtTable and destroy the stringlist.

    if (m_pTable->AddRow(papszCells) == 0)
        hr = S_OK;
    else
    {
        hr = RaiseException(E_FAIL, CPLGetLastErrorMsg());
    }
    CSLDestroy(papszCells);

    return hr;
}

#ifdef __TODO__not_implemented_yet
/**********************************************************************
 *                 OWTCHtTable::SetColumn()
 *
 * nCol is the 0-based column index
 * vabstrData contains the values to set in the cells.
 * Valid input types are BSTR or a SAFEARRAY of BSTR.
 **********************************************************************/
HRESULT __stdcall OWTCHtTable::SetColumn(LONG nCol, VARIANT vabstrData )
{ 
    HRESULT     hr;
    char        **papszCells = NULL;

    trace("SetColumn()") ;

    // The magic of extracting the data from the VARIANT is hidden
    // inside BSTRArrayToStringlist()
    hr = BSTRArrayToStringlist(vabstrData, &papszCells);
    if (FAILED(hr))
        return hr;

    // Pass the cell values to the OWTHtTable and destroy the stringlist.

    if (m_pTable->SetColumn(nCol, papszCells) == 0)
        hr = S_OK;
    else
    {
        hr = RaiseException(E_FAIL, CPLGetLastErrorMsg());
    }
    CSLDestroy(papszCells);

    return hr;
}

/**********************************************************************
 *                 OWTCHtTable::SetRow()
 *
 * nRow is the 0-based row index
 * vabstrData contains the values to set in the cells.
 * Valid input types are BSTR or a SAFEARRAY of BSTR.
 **********************************************************************/
HRESULT __stdcall OWTCHtTable::SetRow(LONG nRow, VARIANT vabstrData  )
{ 
    HRESULT     hr;
    char        **papszCells = NULL;

    trace("SetRow()") ;

    // The magic of extracting the data from the VARIANT is hidden
    // inside BSTRArrayToStringlist()
    hr = BSTRArrayToStringlist(vabstrData, &papszCells);
    if (FAILED(hr))
        return hr;

    // Pass the cell values to the OWTHtTable and destroy the stringlist.

    if (m_pTable->SetRow(nRow, papszCells) == 0)
        hr = S_OK;
    else
    {
        hr = RaiseException(E_FAIL, CPLGetLastErrorMsg());
    }
    CSLDestroy(papszCells);

    return hr;
}
#endif // __TODO__not_implemented_yet

/**********************************************************************
 *                 OWTCHtTable::SetCell()
 *
 * Set the value for all the cells in the specified range.  
 * Create the new cell if necessary and bforceCreate=TRUE (default is FALSE)
 *
 * When passing data for multiple rows, papszData should contain rows
 * of data to cover the whole range.
 *
 * (nCol1, nRow1)-(nCol2, nRow2) are 0-based cell coords. that define
 * the range of cells to set.
 * vabstrData contains the values to set in the cells.
 * Valid input types are BSTR or a SAFEARRAY of BSTR.
 **********************************************************************/
HRESULT __stdcall OWTCHtTable::SetCell(LONG nCol, LONG nRow,
                                       BSTR bstrData, 
                                       VARIANT vbForceCreate)
{ 
    HRESULT     hr;
    char        *pszData = NULL;
    GBool       bForceCreate = FALSE;

    trace("SetCell()") ;

    if (vbForceCreate.vt == VT_BOOL)
        bForceCreate = (vbForceCreate.boolVal == VARIANT_TRUE);

    // Pass the cell value to the OWTHtTable and cleanup.

    hr = UnicodeToAnsi( bstrData, &pszData );
    if( FAILED(hr) )
        return hr;

    if (m_pTable->SetCell(nCol, nRow, pszData, bForceCreate) == 0)
        hr = S_OK;
    else
    {
        hr = RaiseException(E_FAIL, CPLGetLastErrorMsg());
    }
    CoTaskMemFree(pszData);

    return hr;
}

/**********************************************************************
 *                 OWTCHtTable::SetCellRange()
 *
 * Set the value for all the cells in the specified range.  
 * Create the new cell if necessary and bforceCreate=TRUE (default is FALSE)
 *
 * When passing data for multiple rows, papszData should contain rows
 * of data to cover the whole range.
 *
 * (nCol1, nRow1)-(nCol1+numCols-1, nRow1+numRows-1) are 0-based cell 
 * coords. that define the range of cells to set.
 * vabstrData contains the values to set in the cells.
 * Valid input types are BSTR or a SAFEARRAY of BSTR.
 **********************************************************************/
HRESULT __stdcall OWTCHtTable::SetCellRange(LONG nCol1, LONG nRow1,
                                            LONG numCols, LONG numRows,
                                            VARIANT vabstrData, 
                                            VARIANT vbForceCreate)
{ 
    HRESULT     hr;
    char        **papszCells = NULL;
    GBool       bForceCreate = FALSE;

    trace("SetCellRange()") ;

    // The magic of extracting the data from the VARIANT is hidden
    // inside BSTRArrayToStringlist()
    hr = BSTRArrayToStringlist(vabstrData, &papszCells);
    if (FAILED(hr))
        return hr;

    if (vbForceCreate.vt == VT_BOOL)
        bForceCreate = (vbForceCreate.boolVal == VARIANT_TRUE);

    // Pass the cell values to the OWTHtTable and destroy the stringlist.

    if (m_pTable->SetCellRange(nCol1, nRow1, numCols, numRows, 
                               papszCells, bForceCreate) == 0)
        hr = S_OK;
    else
    {
        hr = RaiseException(E_FAIL, CPLGetLastErrorMsg());
    }
    CSLDestroy(papszCells);

    return hr;
}



/**********************************************************************
 *                 OWTCHtTable::SetWrapMaxCols()
 *
 **********************************************************************/
HRESULT __stdcall OWTCHtTable::SetWrapMaxCols(LONG numCols)
{ 
    HRESULT     hr;

    trace("SetWrapMaxCols()") ;

    m_pTable->SetWrapMaxCols(numCols);

    return S_OK;
}


/**********************************************************************
 *                 OWTCHtTable::SetBorderWidth()
 *
 **********************************************************************/
HRESULT __stdcall OWTCHtTable::SetBorderWidth(LONG nBorderWidth)
{ 
    HRESULT     hr;

    trace("SetBorderWidth()") ;

    m_pTable->SetBorderWidth(nBorderWidth);

    return S_OK;
}

/**********************************************************************
 *                 OWTCHtTable::SetReversedOrientation()
 *
 **********************************************************************/
HRESULT __stdcall OWTCHtTable::SetReversedOrientation(VARIANT_BOOL vbRevOrient)
{ 
    HRESULT     hr;

    trace("SetReversedOrientation()") ;

    m_pTable->SetReversedOrientation(vbRevOrient==VARIANT_TRUE);

    return S_OK;
}

/**********************************************************************
 *                 OWTCHtTable::SetNumHeaderRows()
 *
 **********************************************************************/
HRESULT __stdcall OWTCHtTable::SetNumHeaderRows(LONG numRows)
{ 
    HRESULT     hr;

    trace("SetNumHeaderRows()") ;

    m_pTable->SetNumHeaderRows(numRows);

    return S_OK;
}

/**********************************************************************
 *                 OWTCHtTable::SetNumHeaderColumns()
 *
 **********************************************************************/
HRESULT __stdcall OWTCHtTable::SetNumHeaderColumns(LONG numCols)
{ 
    HRESULT     hr;

    trace("SetNumHeaderColumns()") ;

    m_pTable->SetNumHeaderColumns(numCols);

    return S_OK;
}
